import unittest
import jax
import jax.numpy as jnp
import rodeo.kalmantv.square_root as sqrt_ktv
import utils


# --- square-root kalmantv.predict ---------------------------------------------------------


def comparison_check (self, mean1, mean2, var1, var2, places=9):
    """
    Helper function to assert that two values are almost equal with respect to a relative error tolerance.
    """
    self.assertAlmostEqual(utils.rel_err(mean1, mean2),0.0) 
    self.assertAlmostEqual(utils.rel_err(var1, jnp.dot(var2,var2.T)), 0.0, places) 


class TestKalmanTVGM(unittest.TestCase):
    """
    Test if square-root KalmanTV gives the same results as square-root Gaussian Markov process.
    """
    setUp = utils.kalman_setup
       
    def test_predict(self):
        mean_state_past, var_state_past, mean_state_pred1, var_state_pred1, _, _ = utils.test_filter(self)
        mean_state_pred2, var_state_pred2 = sqrt_ktv.predict(
            mean_state_past=mean_state_past,
            var_state_past=jnp.linalg.cholesky(var_state_past),
            mean_state=self.mean_state[1],
            wgt_state=self.wgt_state[0],
            var_state=jnp.linalg.cholesky(self.var_state[1])
        )
        
        comparison_check(self, mean_state_pred1, mean_state_pred2, var_state_pred1, var_state_pred2)

    def test_update(self):
        _, _, mean_state_pred, var_state_pred,  mean_state_filt1, var_state_filt1 = utils.test_filter(self)
        mean_state_filt2, var_state_filt2 = sqrt_ktv.update(
            mean_state_pred=mean_state_pred,
            var_state_pred=jnp.linalg.cholesky(var_state_pred),
            x_meas=self.x_meas[1],
            mean_meas=self.mean_meas[1],
            wgt_meas=self.wgt_meas[1],
            var_meas=jnp.linalg.cholesky(self.var_meas[1])
        )
        
        comparison_check(self, mean_state_filt1, mean_state_filt2, var_state_filt1, var_state_filt2)

    def test_filter(self):
        mean_state_past, var_state_past, mean_state_pred1, \
            var_state_pred1, mean_state_filt1, var_state_filt1 = utils.test_filter(self)
        mean_state_pred2, var_state_pred2, \
            mean_state_filt2, var_state_filt2 = sqrt_ktv.filter(
                mean_state_past=mean_state_past,
                var_state_past=jnp.linalg.cholesky(var_state_past),
                mean_state=self.mean_state[1],
                wgt_state=self.wgt_state[0],
                var_state=jnp.linalg.cholesky(self.var_state[1]),
                x_meas=self.x_meas[1],
                mean_meas=self.mean_meas[1],
                wgt_meas=self.wgt_meas[1],
                var_meas=jnp.linalg.cholesky(self.var_meas[1])
            )

        comparison_check(self, mean_state_pred1, mean_state_pred2, var_state_pred1, var_state_pred2)
        comparison_check(self, mean_state_filt1, mean_state_filt2, var_state_filt1, var_state_filt2)

    def test_filter2(self):
        mean_state_past, var_state_past, mean_state_pred1, var_state_pred1, \
            mean_state_filt1, var_state_filt1 = utils.test_filter2(self)
        
        mean_state_pred2, var_state_pred2, \
            mean_state_filt2, var_state_filt2 = sqrt_ktv.filter(
                mean_state_past=mean_state_past,
                var_state_past=jnp.linalg.cholesky(var_state_past),
                mean_state=self.mean_state[2],
                wgt_state=self.wgt_state[1],
                var_state=jnp.linalg.cholesky(self.var_state[2]),
                x_meas=self.x_meas[2],
                mean_meas=self.mean_meas[2],
                wgt_meas=self.wgt_meas[2],
                var_meas=jnp.linalg.cholesky(self.var_meas[2])
            )

        comparison_check(self, mean_state_pred1, mean_state_pred2, var_state_pred1, var_state_pred2)
        comparison_check(self, mean_state_filt1, mean_state_filt2, var_state_filt1, var_state_filt2)

    def test_smooth_mv(self):
        mean_state_next, var_state_next, mean_state_filt, var_state_filt, mean_state_pred, var_state_pred, \
            mean_state_smooth1, var_state_smooth1, _, _, _, _ = utils.test_smooth(self)
        
        mean_state_smooth2, var_state_smooth2 = sqrt_ktv.smooth_mv(
            mean_state_next=mean_state_next,
            var_state_next=jnp.linalg.cholesky(var_state_next),
            mean_state_filt=mean_state_filt,
            var_state_filt=jnp.linalg.cholesky(var_state_filt),
            mean_state_pred=mean_state_pred,
            var_state_pred=jnp.linalg.cholesky(var_state_pred),
            wgt_state=self.wgt_state[0],
            var_state=jnp.linalg.cholesky(self.var_state[1])
        )

        comparison_check(self, mean_state_smooth1[0], mean_state_smooth2, var_state_smooth1[0, :, 0, :].squeeze(), var_state_smooth2)

    def test_smooth_sim(self):
        _, _, mean_state_filt, var_state_filt, mean_state_pred, var_state_pred, \
            _, _, mean_state_sim1, V, _, _ = utils.test_smooth(self)

        mean_state_sim2, var_state_sim2 = sqrt_ktv.smooth_sim(
            x_state_next=self.x_state_next,
            mean_state_filt=mean_state_filt,
            var_state_filt=jnp.linalg.cholesky(var_state_filt),
            mean_state_pred=mean_state_pred,
            var_state_pred=jnp.linalg.cholesky(var_state_pred),
            wgt_state=self.wgt_state[0],
            var_state=jnp.linalg.cholesky(self.var_state[1])
        )
        
        comparison_check(self, mean_state_sim1, mean_state_sim2, V, var_state_sim2, places=2)

    def test_smooth(self):
        mean_state_next, var_state_next, mean_state_filt, var_state_filt, mean_state_pred, var_state_pred, \
            mean_state_smooth1, var_state_smooth1, mean_state_sim1, V, _, _ = utils.test_smooth(self)

        mean_state_sim2, var_state_sim2, mean_state_smooth2, var_state_smooth2 = sqrt_ktv.smooth(
            x_state_next=self.x_state_next,
            mean_state_next=mean_state_next,
            var_state_next=jnp.linalg.cholesky(var_state_next),
            mean_state_filt=mean_state_filt,
            var_state_filt=jnp.linalg.cholesky(var_state_filt),
            mean_state_pred=mean_state_pred,
            var_state_pred=jnp.linalg.cholesky(var_state_pred),
            wgt_state=self.wgt_state[0],
            var_state=jnp.linalg.cholesky(self.var_state[1])
        )
        
        comparison_check(self, mean_state_sim1, mean_state_sim2, V, var_state_sim2, places=2)
        comparison_check(self, mean_state_smooth1[0], mean_state_smooth2, var_state_smooth1[0, :, 0, :].squeeze(), var_state_smooth2)

    def test_smooth_cond(self):
        _, _, mean_state_filt, var_state_filt, mean_state_pred, \
            var_state_pred, _, _, _, V, A, b =  utils.test_smooth(self)
        # mean_state_sim1 = A.dot(self.x_state_next)+b
        # x_state_smooth1 = random.multivariate_normal(self.key, A.dot(self.x_state_next)+b, V)
        # x_state_smooth1 = ktv._state_sim(
        #    mean_state=A.dot(self.x_state_next)+b,
        #    var_state=V,
        #    z_state=self.z_state
        # )

        A2, b2, V2 = jax.jit(sqrt_ktv.smooth_cond)(
            mean_state_filt=mean_state_filt,
            var_state_filt=jnp.linalg.cholesky(var_state_filt),
            mean_state_pred=mean_state_pred,
            var_state_pred=jnp.linalg.cholesky(var_state_pred),
            wgt_state=self.wgt_state[0],
            var_state=jnp.linalg.cholesky(self.var_state[1])
        )

        self.assertAlmostEqual(utils.rel_err(A, A2), 0.0)
        comparison_check(self, b, b2, V, V2, places=2)

   
if __name__ == '__main__':
    unittest.main()
    
    



import unittest
import jax
import jax.numpy as jnp
import jax.scipy as jsp
import jax.random as random
from rodeo.solve import *
from rodeo.interrogate import *
import utils
# from jax import config
# config.update("jax_enable_x64", True)
    
class TestrodeoJit(unittest.TestCase):
    """
    Check whether jit and unjitted gives the same result.
    
    """
    setUp = utils.fitz_setup

    def test_interrogate_rodeo(self):
        # without jit
        wgt_meas1, mean_meas1, var_meas1 = interrogate_rodeo(
            self.key,
            self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R
        )
        # with jit
        rodeo_jit = jax.jit(interrogate_rodeo, static_argnums=(1,))
        wgt_meas2, mean_meas2, var_meas2 = rodeo_jit(
            self.key,
            self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R
        )
        # objective function for gradient
        def obj_fun(theta):
            return jnp.mean(
                interrogate_rodeo(
                    self.key, self.fitz_jax,
                    ode_weight=self.W_block, t=self.t, theta=theta,
                    mean_state_pred=self.x0_block,
                    var_state_pred=self.prior_R)[0])
        # grad without jit
        grad1 = jax.grad(obj_fun)(self.theta)
        # grad with jit
        grad2 = jax.jit(jax.grad(obj_fun))(self.theta)
        self.assertAlmostEqual(utils.rel_err(wgt_meas1, wgt_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(mean_meas1, mean_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(var_meas1, var_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(grad1, grad2), 0.0)

    def test_interrogate_chkrebtii(self):
        # without jit
        wgt_meas1, mean_meas1, var_meas1 = interrogate_chkrebtii(
            self.key,
            self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R,
            kalman_type="standard"
        )
        # with jit
        rodeo_jit = jax.jit(interrogate_chkrebtii, static_argnums=(1,6))
        wgt_meas2, mean_meas2, var_meas2 = rodeo_jit(
            self.key,
            self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R,
            kalman_type="standard"
        )
        # objective function for gradient
        def obj_fun(theta):
            return jnp.mean(
                interrogate_chkrebtii(
                    self.key, self.fitz_jax,
                    ode_weight=self.W_block, t=self.t, theta=theta,
                    mean_state_pred=self.x0_block,
                    var_state_pred=self.prior_R,
                    kalman_type="standard")[0])
        # grad without jit
        grad1 = jax.grad(obj_fun)(self.theta)
        # grad with jit
        grad2 = jax.jit(jax.grad(obj_fun))(self.theta)
        self.assertAlmostEqual(utils.rel_err(wgt_meas1, wgt_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(mean_meas1, mean_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(var_meas1, var_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(grad1, grad2), 0.0)

    def test_interrogate_schober(self):
        # without jit
        wgt_meas1, mean_meas1, var_meas1 = interrogate_schober(
            self.key,
            self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R
        )
        # with jit
        rodeo_jit = jax.jit(interrogate_schober, static_argnums=(1,))
        wgt_meas2, mean_meas2, var_meas2 = rodeo_jit(
            self.key,
            self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R
        )
        # objective function for gradient
        def obj_fun(theta):
            return jnp.mean(
                interrogate_schober(
                    self.key, self.fitz_jax,
                    t=self.t, theta=theta,
                    ode_weight=self.W_block,
                    mean_state_pred=self.x0_block,
                    var_state_pred=self.prior_R)[0])
        # grad without jit
        grad1 = jax.grad(obj_fun)(self.theta)
        # grad with jit
        grad2 = jax.jit(jax.grad(obj_fun))(self.theta)
        self.assertAlmostEqual(utils.rel_err(wgt_meas1, wgt_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(mean_meas1, mean_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(var_meas1, var_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(grad1, grad2), 0.0)

    def test_mv(self):
        # without jit
        mu1, var1 = solve_mv(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                             ode_init=self.x0_block, theta=self.theta,
                             t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                             prior_pars=self.prior_pars,
                             interrogate=interrogate_rodeo)
        # with jit
        mv_jit = jax.jit(solve_mv, static_argnums=(1, 6, 7))
        mu2, var2 = mv_jit(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                            ode_init=self.x0_block, theta=self.theta,
                            t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                            prior_pars=self.prior_pars,
                            interrogate=interrogate_rodeo)
        # objective function for gradient
        def obj_fun(theta):
            return jnp.mean(
                solve_mv(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                        ode_init=self.x0_block, theta=self.theta,
                        t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                        prior_pars=self.prior_pars,
                        interrogate=interrogate_rodeo)[0])
        # grad without jit
        grad1 = jax.grad(obj_fun)(self.theta)
        # grad with jit
        grad2 = jax.jit(jax.grad(obj_fun))(self.theta)
        self.assertAlmostEqual(utils.rel_err(mu1, mu2), 0.0)
        self.assertAlmostEqual(utils.rel_err(var1, var2), 0.0)
        self.assertAlmostEqual(utils.rel_err(grad1, grad2), 0.0)
    
    def test_sim(self):
        # without jit
        sim1 = solve_sim(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                        ode_init=self.x0_block, theta=self.theta,
                        t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                        prior_pars=self.prior_pars,
                        interrogate=interrogate_rodeo)
        # with jit
        sim_jit = jax.jit(solve_sim, static_argnums=(1, 6, 7))
        sim2 = sim_jit(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                       ode_init=self.x0_block, theta=self.theta,
                       t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                       prior_pars=self.prior_pars,
                       interrogate=interrogate_rodeo)
        # objective function for gradient
        def obj_fun(theta):
            return jnp.mean(
                solve_sim(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                        ode_init=self.x0_block, theta=self.theta,
                        t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                        prior_pars=self.prior_pars,
                        interrogate=interrogate_rodeo)[0])
        # grad without jit
        grad1 = jax.grad(obj_fun)(self.theta)
        # grad with jit
        grad2 = jax.jit(jax.grad(obj_fun))(self.theta)
        self.assertAlmostEqual(utils.rel_err(sim1, sim2), 0.0)
        self.assertAlmostEqual(utils.rel_err(grad1, grad2), 0.0)
    
    # def test_solve(self):
    #     # without jit
    #     sim1, mu1, var1 = \
    #         solve(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
    #               ode_init=self.x0_block, theta=self.theta,
    #               t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
    #               prior_weight=self.prior_Q, prior_var=self.prior_R,
    #               interrogate=interrogate_rodeo)
    #     # with jit
    #     solve_jit = jax.jit(solve, static_argnums=(1, 6, 7))
    #     sim2, mu2, var2 = \
    #         solve_jit(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
    #                   ode_init=self.x0_block, theta=self.theta,
    #                   t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
    #                   prior_weight=self.prior_Q, prior_var=self.prior_R,
    #                   interrogate=interrogate_rodeo)
    #     # objective function for gradient
    #     def obj_fun(theta):
    #         return jnp.mean(
    #             solve(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
    #                   ode_init=self.x0_block, theta=self.theta,
    #                   t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
    #                   prior_weight=self.prior_Q, prior_var=self.prior_R,
    #                   interrogate=interrogate_rodeo)[0])
    #     # grad without jit
    #     grad1 = jax.grad(obj_fun)(self.theta)
    #     # grad with jit
    #     grad2 = jax.jit(jax.grad(obj_fun))(self.theta)
    #     self.assertAlmostEqual(utils.rel_err(sim1, sim2), 0.0)
    #     self.assertAlmostEqual(utils.rel_err(mu1, mu2), 0.0)
    #     self.assertAlmostEqual(utils.rel_err(var1, var2), 0.0)
    #     self.assertAlmostEqual(utils.rel_err(grad1, grad2), 0.0)

if __name__ == '__main__':
    unittest.main()



r"""
Util functions for rodeo.

"""
import jax
import jax.numpy as jnp
import jax.scipy as jsp


def add_sqrt(sqrt_A, sqrt_B):
    r"""
    Transforms the square roots of matrices A and B into the square root of their sum.

    Args:
        sqrt_A (ndarray(n_dim, n_dim)): The square root of matrix A.
        sqrt_B (ndarray(n_dim, n_dim)): The square root of matrix B.

    Returns:
        (ndarray(n_dim, n_dim)): The square root of the sum of matrices A and B.
    """

    sqrt_sum = jnp.vstack([sqrt_A.T,sqrt_B.T])
    Q,R = jnp.linalg.qr(sqrt_sum)
    return R.T


def mvncond(mu, Sigma, icond):
    """
    Calculates A, b, and V such that :math:`y[!icond] | y[icond] \sim \operatorname{Normal}(A y[icond] + b, V)`.

    Args:
        mu (ndarray(2*n_dim)): Mean of y.
        Sigma (ndarray(2*n_dim, 2*n_dim)): Covariance of y. 
        icond (ndarray(2*nd_dim)): Conditioning on the terms given.

    Returns:
        (tuple):
        - **A** (ndarray(n_dim, n_dim)): For :math:`y \sim \operatorname{Normal}(\mu, \Sigma)` 
          such that :math:`y[!icond] | y[icond] \sim \operatorname{Normal}(A y[icond] + b, V)` Calculate A.
        - **b** (ndarray(n_dim)): For :math:`y \sim \operatorname{Normal}(\mu, \Sigma)` 
          such that :math:`y[!icond] | y[icond] \sim \operatorname{Normal}(A y[icond] + b, V)` Calculate b.
        - **V** (ndarray(n_dim, n_dim)): For :math:`y \sim \operatorname{Normal}(\mu, \Sigma)`
          such that :math:`y[!icond] | y[icond] \sim \operatorname{Normal}(A y[icond] + b, V)` Calculate V.

    """
    # if y1 = y[~icond] and y2 = y[icond], should have A = Sigma12 * Sigma22^{-1}
    ficond = jnp.nonzero(~icond)
    ticond = jnp.nonzero(icond)
    # A = jnp.dot(Sigma[jnp.ix_(ficond[0], ticond[0])], jsp.cho_solve(
    #     jsp.cho_factor(Sigma[jnp.ix_(ticond[0], ticond[0])]), jnp.identity(sum(icond))))
    A = jnp.dot(Sigma[jnp.ix_(ficond[0], ticond[0])],
                solve_var(Sigma[jnp.ix_(ticond[0], ticond[0])],
                          jnp.identity(jnp.sum(icond)))
                )
    b = mu[~icond] - jnp.dot(A, mu[icond])  # mu1 - A * mu2
    V = Sigma[jnp.ix_(ficond[0], ficond[0])] - jnp.dot(A, Sigma[jnp.ix_(ticond[0], ficond[0])])  # Sigma11 - A * Sigma21
    return A, b, V


def multivariate_normal_logpdf(x, mean, cov):
    r"""Using eigendecomposition to compute multivariate normal logpdf.
    
    Args:
        x (ndarray(p)): Observations.
        mean (ndarray(p)): Mean of the distribution.
        cov (ndarray(p, p)): Symmetric positive (semi)definite covariance matrix of the distribution.
    
    Returns:
        (float): The logpdf of the multivariate normal.
    """
    w, v = jnp.linalg.eigh(cov)
    z = jnp.dot(v.T, x - mean)
    z2 = z**2
    iw = ~jnp.isclose(w, 0, rtol=1e-300)
    w = jnp.where(iw, w, 1.) # remove possibility of nan
    val = z2/w + jnp.log(w)
    val = -.5 * jnp.sum(jnp.where(iw, val, 0.)) - jnp.sum(iw)*.5*jnp.log(2*jnp.pi) 
    return val

def first_order_pad(ode_fun, n_vars, n_deriv):
    r"""
    Returns the W matrix, and a function for finding the initial value
    for given :math:`\theta`.

    Args:
        ode_fun (Callable): ODE function.
        n_vars (int): Number of variables.
        n_deriv (int): Number of upper derivatives to use.
        
    Returns:
        (tuple):
        - **W** (ndarray(n_var, 1, n_deriv)): W matrix defining the left hand side of the ODE.
        - **ode_init** (Callable): A function that helps setting up the initial state space mean.
    """
    
    def ode_init(x0, t, **params):
        x0 = x0[:, None]
        return jnp.hstack([x0, ode_fun(x0, t, **params), jnp.zeros((n_vars, n_deriv-2))])
    
    W = jnp.zeros((n_vars, 1, n_deriv))
    W = W.at[:, :, 1].set(1.0)
    return W, ode_init


def solve_var(V, B):
    r"""
    Computes :math:`X = V^{-1}B`, where :math:`V` is a variance matrix.

    Args:
        V (ndarray(n_dim1, n_dim1)): Variance matrix :math:`V`.
        B (ndarray(n_dim1, n_dim2)): Matrix :math:`B`.

    Returns:
        (ndarray(n_dim1, n_dim2)): Matrix :math:`X = V^{-1}B`.
    """

    # L, low = jsp.linalg.cho_factor(V)
    # return jsp.linalg.cho_solve((L, low), B)
    return jnp.linalg.solve(V, B)


r"""
This module implements the Fenrir algorithm as described in Tronarp et al 2022 for computing the approximate likelihood of :math:`p(Y_{0:M} \mid Z_{1:N})`.

The forward pass model is

.. math::

    x_0 = v

    X_n = c_n + Q_n X_{n-1} + R_n^{1/2} \epsilon_n

    Z_n = W_n X_n - f(X_n, t_n) + V_n^{1/2} \eta_n.

We assume that :math:`c_n = 0, Q_n = Q, R_n = R`, and :math:`W_n = W` for all :math:`n`. Using the Kalman filtering recursions, the above model can be simulated via the reverse pass model

.. math::

    X_N \sim \operatorname{Normal}(b_N, C_N)

    X_n = A_n X_{n+1} + b_n + C_n^{1/2} \epsilon_n.
    
Fenrir combines the observations

.. math::

    Y_m = D_m X_m + \Omega^{1/2}_m \eta_m,

with the reverse pass model to condition on data. Here :math:`\epsilon_n, \eta_m` are standard normals.
"""
import jax
import jax.numpy as jnp
from rodeo.kalmantv import standard
from rodeo.kalmantv import square_root
from rodeo.solve import _solve_filter
from rodeo.utils import multivariate_normal_logpdf

# --- helper functions --------------------------------------------------------


def _forecast_update(mean_state_pred, var_state_pred,
                     x_meas, mean_meas,
                     wgt_meas, var_meas,
                     kalman_funs):
    r"""
    Perform one update step of the Kalman filter and forecast.

    Args:
        mean_state_pred (ndarray(n_block, n_bstate)): Mean estimate for state at time n given observations from times [0...n-1]; denoted by :math:`\mu_{n|n-1}`.
        var_state_pred (ndarray(n_block, n_bstate, n_sbtate)): Covariance of estimate for state at time n given observations from times [0...n-1]; denoted by :math:`\Sigma_{n|n-1}`.
        x_meas (ndarray(n_block, n_bmeas)): Interrogated measure vector from `x_state`; :math:`y_n`.
        mean_meas (ndarray(n_block, n_bmeas)): Transition offsets defining the measure prior.
        wgt_meas (ndarray(n_block, n_bmeas, n_bstate)): Transition matrix defining the measure prior.
        var_meas (ndarray(n_block, n_bmeas, n_bmeas)): Variance matrix defining the measure prior.
        kalman_funs (object): An object that contains the Kalman filtering functions: predict, update and smooth.

    Returns:
        (tuple):
        - **logdens** (float): The loglikelihood for the observations.
        - **mean_state_filt** (ndarray(n_block, n_bstate)): Mean estimate for state at time n given observations from times [0...n]; denoted by :math:`\mu_{n|n}`.
        - **var_state_filt** (ndarray(n_block, n_bstate, n_bstate)): Covariance of estimate for state at time n given observations from times [0...n]; denoted by :math:`\Sigma_{n|n}`.
    """
    # kalman forecast
    mean_state_fore, var_state_fore = kalman_funs.forecast(
        mean_state_pred=mean_state_pred,
        var_state_pred=var_state_pred,
        mean_meas=mean_meas,
        wgt_meas=wgt_meas,
        var_meas=var_meas
    )
    logdens = multivariate_normal_logpdf(
        x_meas, mean=mean_state_fore, cov=var_state_fore)
    # kalman update
    mean_state_filt, var_state_filt = kalman_funs.update(
        mean_state_pred=mean_state_pred,
        var_state_pred=var_state_pred,
        x_meas=x_meas,
        mean_meas=mean_meas,
        wgt_meas=wgt_meas,
        var_meas=var_meas
    )
    return logdens, mean_state_filt, var_state_filt

# --- loglikelihood -----------------------------------------------------------


def _backward(mean_state_filt, var_state_filt,
              mean_state_pred, var_state_pred,
              prior_weight, prior_var,
              t_min, t_max, n_steps,
              obs_data, obs_times,
              obs_weight, obs_var,
              kalman_funs):
    r"""
    Compute the backward Markov chain parameters and forward pass but backwards in time.

    Args:
        mean_state_pred (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time n given observations from times [0...n]; denoted by :math:`\mu_{n|n-1}`.
        var_state_pred (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Covariance of estimate for state at time n given observations from times [0...n-1]; denoted by :math:`\Sigma_{n|n-1}`.
        mean_state_filt (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time n given observations from times [0...n]; denoted by :math:`\mu_{n|n}`.
        var_state_filt (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Covariance of estimate for state at time n given observations from times [0...n]; denoted by :math:`\Sigma_{n|n}`.
        prior_weight (ndarray(n_block, n_bstate, n_bstate)): Weight matrix defining the solution prior; :math:`Q`.
        prior_var (ndarray(n_block, n_bstate, n_bstate)): Variance matrix defining the solution prior; :math:`R`.
        t_min (float): First time point of the time interval to be evaluated; :math:`t_0`.
        t_max (float): Last time point of the time interval to be evaluated; :math:`t_N`.
        n_steps (int): Number of discretization points (:math:`N`) of the time interval that is evaluated, such that discretization timestep is :math:`dt = (b-a)/N`.
        obs_data (ndarray(n_obs, n_blocks, n_bobs)): Observed data; :math:`y_{0:M}`.
        obs_times (ndarray(n_obs)): Observation time; :math:`0, \ldots, M`.
        obs_weight (ndarray(n_obs, n_blocks, n_bobs, n_bstate)): Weight matrix in the observation model; :math:`D_{0:M}`.
        obs_var (ndarry(n_obs, n_blocks, n_bobs, n_bobs)): Variance matrix in the observation model; :math:`\Omega_{0:M}`
        kalman_funs (object): An object that contains the Kalman filtering functions: predict, update and smooth.

    Returns:
        (float) : The logdensity of :math:`p(y_{0:M} \mid Z_{1:N})`.

    """
    # get dimensions
    n_obs, n_block, n_bobs, n_bstate = obs_weight.shape
    # insert observations on solver time grid
    sim_times = jnp.linspace(t_min, t_max, n_steps + 1)
    obs_ind = jnp.searchsorted(sim_times, obs_times)
    # offset of obs is assumed to be 0
    obs_mean = jnp.zeros((n_block, n_bobs))
    # forecast function without kalman_funs
    forecast_update = lambda mean_state_pred, var_state_pred,\
                             x_meas, mean_meas, wgt_meas, var_meas\
                             : _forecast_update(mean_state_pred, var_state_pred,
                                                x_meas, mean_meas,
                                                wgt_meas, var_meas,
                                                kalman_funs)

    def scan_fun(carry, forward_states):
        # Kalman filter backwards in time
        bmean_state_filt, bvar_state_filt = carry["state_filt"] 
        # Kalman filter estimates from forward
        mean_state_filt, var_state_filt = forward_states["state_filt"]
        mean_state_pred, var_state_pred = forward_states["state_pred"]

        logdens = carry["logdens"]
        i = carry["i"]
        t = forward_states["t"] # t_n
        # get Markov params
        wgt_state_back, mean_state_back, var_state_back = jax.vmap(kalman_funs.smooth_cond)(
                mean_state_filt=mean_state_filt,
                var_state_filt=var_state_filt,
                mean_state_pred=mean_state_pred,
                var_state_pred=var_state_pred,
                wgt_state=prior_weight,
                var_state=prior_var
        )
        # kalman predict
        bmean_state_pred, bvar_state_pred = jax.vmap(kalman_funs.predict)(
                mean_state_past=bmean_state_filt,
                var_state_past=bvar_state_filt,
                mean_state=mean_state_back,
                wgt_state=wgt_state_back,
                var_state=var_state_back
        )

        # not t time point of observation
        def _no_obs():
            return bmean_state_pred, bvar_state_pred, 0.0, i

        
        # at time point of observation
        def _obs():
            # kalman forecast and update
            logp, bmean_state_next, bvar_state_next = jax.vmap(forecast_update)(
                    mean_state_pred=bmean_state_pred,
                    var_state_pred=bvar_state_pred,
                    x_meas=obs_data[i],
                    mean_meas=obs_mean,
                    wgt_meas=obs_weight[i],
                    var_meas=obs_var[i]
            )
            return bmean_state_next, bvar_state_next, jnp.sum(logp), i-1

        bmean_state_filt, bvar_state_filt, logp, i = jax.lax.cond(
            obs_ind[i] == t, _obs, _no_obs)
        logdens += logp

        # output
        carry = {
            "state_filt": (bmean_state_filt, bvar_state_filt),
            "logdens": logdens,
            "i": i
        }
        stack = {
            "state_pred": (bmean_state_pred, bvar_state_pred),
            "state_filt": (bmean_state_filt, bvar_state_filt),
            "wgt_state": wgt_state_back,
            "var_state": var_state_back 
        }
        return carry, stack

    # terminal point update
    mean_state_term = mean_state_filt[n_steps]
    var_state_term = var_state_filt[n_steps]
    logdens = 0.0
    i = n_obs - 1
    # no observations
    def _no_obs():
        # no need to update
        return mean_state_term, var_state_term, 0.0, i

    # observation
    def _obs():
        # kalman forecast and update
        logp, bmean_state_next, bvar_state_next = jax.vmap(forecast_update)(
                mean_state_pred=mean_state_term,
                var_state_pred=var_state_term,
                x_meas=obs_data[i],
                mean_meas=obs_mean,
                wgt_meas=obs_weight[i],
                var_meas=obs_var[i]
        )
        return bmean_state_next, bvar_state_next, jnp.sum(logp), i-1

    bmean_state_filt, bvar_state_filt, logp, i = jax.lax.cond(
        obs_ind[i] >= n_steps, _obs, _no_obs)
    logdens += logp

    # start at N 
    scan_init = {
        "state_filt": (bmean_state_filt, bvar_state_filt),
        "logdens": logdens,
        "i": i
    }
    forward_states_init = {
        "state_pred": (mean_state_pred[1:n_steps+1], var_state_pred[1:n_steps+1]),
        "state_filt": (mean_state_filt[:n_steps], var_state_filt[:n_steps]),
        "t": jnp.arange(n_steps)
    }

    scan_out, scan_out2 = jax.lax.scan(
        scan_fun, scan_init, forward_states_init, reverse=True)

    # append initial values to back
    mean_scan_pred, var_scan_pred = scan_out2["state_pred"]
    mean_scan_filt, var_scan_filt = scan_out2["state_filt"]
    mean_state_pred = jnp.concatenate(
        [mean_scan_pred, mean_state_term[None]]
    )
    var_state_pred = jnp.concatenate(
        [var_scan_pred, var_state_term[None]]
    )
    mean_state_filt = jnp.concatenate(
        [mean_scan_filt, bmean_state_filt[None]]
    )
    var_state_filt = jnp.concatenate(
        [var_scan_filt, bvar_state_filt[None]]
    )
    # repack
    scan_out2 = {
        "state_pred": (mean_state_pred, var_state_pred),
        "state_filt": (mean_state_filt, var_state_filt),
        "wgt_state": scan_out2["wgt_state"],
        "var_state": scan_out2["var_state"]
    }
    return scan_out["logdens"], scan_out2

def fenrir(key, ode_fun, ode_weight, ode_init,
           t_min, t_max, n_steps,
           interrogate,
           prior_pars,
           obs_data, obs_times, obs_weight, obs_var,
           kalman_type="standard", **params):
    r"""
    Fenrir algorithm to compute the approximate loglikelihood of :math:`p(Y_{0:M} \mid Z_{1:N})`.

    Args:
        key (PRNGKey): PRNG key.
        ode_fun (Callable): Higher order ODE function :math:`W X_t = F(X_t, t)` taking arguments :math:`X` and :math:`t`.
        ode_weight (ndarray(n_block, n_bmeas, n_bstate)): Weight matrix defining the measure prior; :math:`W`.
        ode_init (ndarray(n_block, n_bstate)): Initial value of the state variable :math:`X_t` at time :math:`t = a`.
        t_min (float): First time point of the time interval to be evaluated; :math:`a`.
        t_max (float): Last time point of the time interval to be evaluated; :math:`b`.
        n_steps (int): Number of discretization points (:math:`N`) of the time interval that is evaluated, such that discretization timestep is :math:`dt = (b-a)/N`.
        interrogate (Callable): Function defining the interrogation method.
        prior_pars (tuple): A tuple containing the weight matrix and the variance matrix defining the solution prior; :math:`Q, R`.
        obs_data (ndarray(n_obs, n_blocks, n_bobs)): Observed data; :math:`y_{0:M}`.
        obs_times (ndarray(n_obs)): Observation time; :math:`0, \ldots, M`.
        obs_weight (ndarray(n_obs, n_blocks, n_bobs, n_bstate)): Weight matrix in the observation model; :math:`D_{0:M}`.
        obs_var (ndarry(n_obs, n_blocks, n_bobs, n_bobs)): Variance matrix in the observation model; :math:`\Omega_{0:M}`
        kalman_type (str): Determine which type of Kalman (standard, square-root) to use.
        params (kwargs): Optional model parameters.

    Returns:
        (float) : The loglikelihood of :math:`p(y_{0:M} \mid Z_{1:N})`.

    """
    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError

    # prior variables
    prior_weight, prior_var = prior_pars

    # forward pass
    filt_out = _solve_filter(
        key=key,
        ode_fun=ode_fun, ode_weight=ode_weight, ode_init=ode_init,
        t_min=t_min, t_max=t_max, n_steps=n_steps,
        interrogate=interrogate,
        prior_weight=prior_weight, prior_var=prior_var,
        kalman_funs=kalman_funs, **params
    )
    mean_state_pred, var_state_pred = filt_out["state_pred"]
    mean_state_filt, var_state_filt = filt_out["state_filt"]

    # backward pass
    logdens, _ = _backward(
        mean_state_filt=mean_state_filt,
        var_state_filt=var_state_filt,
        mean_state_pred=mean_state_pred,
        var_state_pred=var_state_pred,
        prior_weight=prior_weight,
        prior_var=prior_var,
        t_min=t_min, t_max=t_max, n_steps=n_steps,
        obs_data=obs_data, obs_times=obs_times,
        obs_weight=obs_weight, obs_var=obs_var,
        kalman_funs=kalman_funs
    )

    return logdens

# --- ODE solver --------------------------------------------------------------


def _smooth_mv(state_par, kalman_funs):
    r"""
    Smoothing pass of the Fenrir algorithm used to compute solution posterior.

    Args:
        state_par (dict): Dictionary containing the weight, mean and variance matrices of the predicted and updated steps of the Kalman filter.
        kalman_funs (object): An object that contains the Kalman filtering functions: predict, update and smooth.


    Returns:
        (tuple):
        - **mean_state_smooth** (ndarray(n_steps+1, n_block, n_bstate)): Posterior mean of the solution process :math:`X_t` at times :math:`t \in [a, b]`.
        - **var_state_smooth** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Posterior variance of the solution process at times :math:`t \in [a, b]`.

    """
    mean_state_pred, var_state_pred = state_par["state_pred"]
    mean_state_filt, var_state_filt = state_par["state_filt"]
    wgt_state = state_par["wgt_state"]
    var_state = state_par["var_state"]
    n_tot = mean_state_pred.shape[0]
    # smooth pass
    # lax.scan setup

    def scan_fun(state_next, smooth_kwargs):
        mean_state_filt = smooth_kwargs['mean_state_filt']
        var_state_filt = smooth_kwargs['var_state_filt']
        mean_state_pred = smooth_kwargs['mean_state_pred']
        var_state_pred = smooth_kwargs['var_state_pred']
        wgt_state = smooth_kwargs['wgt_state']
        var_state = smooth_kwargs['var_state']
        mean_state_curr, var_state_curr = jax.vmap(kalman_funs.smooth_mv)(
                mean_state_next=state_next["mean"],
                var_state_next=state_next["var"],
                wgt_state=wgt_state,
                mean_state_filt=mean_state_filt,
                var_state_filt=var_state_filt,
                mean_state_pred=mean_state_pred,
                var_state_pred=var_state_pred,
                var_state=var_state
        )
        state_curr = {
            "mean": mean_state_curr,
            "var": var_state_curr
        }
        return state_curr, state_curr
    # initialize
    scan_init = {
        "mean": mean_state_filt[1],
        "var": var_state_filt[1]
    }
    # scan arguments
    scan_kwargs = {
        'mean_state_filt': mean_state_filt[2:],
        'var_state_filt': var_state_filt[2:],
        'mean_state_pred': mean_state_pred[1:n_tot-1],
        'var_state_pred': var_state_pred[1:n_tot-1],
        'wgt_state': wgt_state[1:n_tot],
        'var_state': var_state[1:n_tot]
    }
    # Note: initial value x0 is assumed to be known, so no need to smooth it
    _, scan_out = jax.lax.scan(scan_fun, scan_init, scan_kwargs)

    # append initial values to front and back
    mean_state_smooth = jnp.concatenate(
        [mean_state_filt[0:2], scan_out["mean"]]
    )
    var_state_smooth = jnp.concatenate(
        [var_state_filt[0:2], scan_out["var"]]
    )
    return mean_state_smooth, var_state_smooth


def solve_mv(key, ode_fun, ode_weight, ode_init,
             t_min, t_max, n_steps,
             interrogate,
             prior_pars,
             obs_data, obs_times, obs_weight, obs_var,
             kalman_type="standard", **params):
    r"""
    Fenrir algorithm to compute the mean and variance of :math:`p(X_{0:N} \mid Z_{1:N}, Y_{0:M})`. Same arguments as :func:`fenrir`.

    Returns:
        (tuple):
        - **mean_state_smooth** (ndarray(n_steps+1, n_block, n_bstate)): Posterior mean of the solution process :math:`X_t` at times :math:`t \in [a, b]`.
        - **var_state_smooth** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Posterior variance of the solution process at times :math:`t \in [a, b]`.

    """
    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError
    
    prior_weight, prior_var = prior_pars

    # forward pass
    filt_out = _solve_filter(
        key=key,
        ode_fun=ode_fun, ode_weight=ode_weight, ode_init=ode_init,
        t_min=t_min, t_max=t_max, n_steps=n_steps,
        interrogate=interrogate,
        prior_weight=prior_weight, prior_var=prior_var,
        kalman_funs=kalman_funs, **params
    )
    mean_state_pred, var_state_pred = filt_out["state_pred"]
    mean_state_filt, var_state_filt = filt_out["state_filt"]

    # backward pass
    _, state_par = _backward(
        mean_state_filt=mean_state_filt,
        var_state_filt=var_state_filt,
        mean_state_pred=mean_state_pred,
        var_state_pred=var_state_pred,
        prior_weight=prior_weight,
        prior_var=prior_var,
        t_min=t_min, t_max=t_max, n_steps=n_steps,
        obs_data=obs_data, obs_times=obs_times,
        obs_weight=obs_weight, obs_var=obs_var,
        kalman_funs=kalman_funs
    )

    mean_state_smooth, var_state_smooth = _smooth_mv(state_par, kalman_funs)
    return mean_state_smooth, var_state_smooth



r"""
Stochastic block solver for ODE initial value problems.

The ODE-IVP to be solved is defined as

.. math:: W X_t = f(X_t, t, \theta)

on the time interval :math:`t \in [a, b]` with initial condition :math:`X_a = x_0`.  In the paper, we use the notation :math:`a = t_{\mathrm{min}}` and :math:`b = t_{\mathrm{max}}`.

The stochastic solver proceeds via Kalman filtering and smoothing of "interrogations" of the ODE model as described in Chkrebtii et al 2016, Schober et al 2019.  In the context of the underlying Kalman filterer/smoother, the Gaussian state-space model is

.. math::

    X_n = c_n + Q_n x_{n-1} + R_n^{1/2} \epsilon_n

    Z_n = W_n X_n - f(X_n, t, \theta) + V_n^{1/2} \eta_n.

We assume that :math:`c_n = c, Q_n = Q, R_n = R`, and :math:`W_n = W` for all :math:`n`.

This module optimizes the calculations when :math:`Q`, :math:`R`, and :math:`W`, are block diagonal matrices of conformable and "stackable" sizes.  That is, recall that the dimension of these matrices are `n_state x n_state`, `n_state x n_state`, and `n_meas x n_state`, respectively.  Then suppose that :math:`Q` and :math:`R` consist of `n_block` blocks of size `n_bstate x n_bstate`, where `n_bstate = n_state/n_block`, and :math:`W` consists of `n_block` blocks of size `n_bmeas x n_bstate`, where `n_bmeas = n_meas/n_block`.  Then :math:`Q`, :math:`R`, :math:`W` can be stored as 3D arrays of size `n_block x n_bstate x n_bstate` and `n_block x n_bmeas x n_bstate`.  It is under this paradigm that the `ode` module operates.

"""


import jax
import jax.numpy as jnp
from rodeo.kalmantv import standard
from rodeo.kalmantv import square_root


def _solve_filter(key, ode_fun, ode_weight, ode_init,
                  t_min, t_max, n_steps,
                  interrogate,
                  prior_weight, prior_var,
                  kalman_funs, **params):
    r"""
    Forward pass of the ODE solver. Same arguments as :func:`~ode.solve_mv`.

    Returns:
        (tuple):
        - **mean_state_pred** (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time t given observations from times [a...t-1] for :math:`t \in [a, b]`.
        - **var_state_pred** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Variance estimate for state at time t given observations from times [a...t-1] for :math:`t \in [a, b]`.
        - **mean_state_filt** (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time t given observations from times [a...t] for :math:`t \in [a, b]`.
        - **var_state_filt** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Variance estimate for state at time t given observations from times [a...t] for :math:`t \in [a, b]`.

    """
    # Dimensions of block, state and measure variables
    n_block, n_bmeas, n_bstate = ode_weight.shape

    # arguments for kalman_filter and kalman_smooth
    x_meas = jnp.zeros((n_block, n_bmeas))
    mean_state = jnp.zeros((n_block, n_bstate))
    mean_state_init = ode_init
    var_state_init = jnp.zeros((n_block, n_bstate, n_bstate))

    # forward pass
    def scan_fun(carry, filter_kwargs):
        mean_state_filt, var_state_filt = carry["state_filt"]
        t = filter_kwargs["t"]
        key = filter_kwargs["key"]
        # kalman predict
        mean_state_pred, var_state_pred = jax.vmap(kalman_funs.predict)(
                mean_state_past=mean_state_filt,
                var_state_past=var_state_filt,
                mean_state=mean_state,
                wgt_state=prior_weight,
                var_state=prior_var
        )
        # model interrogation
        wgt_meas, mean_meas, var_meas = interrogate(
            key=key,
            ode_fun=ode_fun,
            ode_weight=ode_weight,
            t=t_min + (t_max-t_min)*(t+1)/n_steps,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            **params
        )
        W_meas = ode_weight + wgt_meas
        # kalman update
        mean_state_next, var_state_next = jax.vmap(kalman_funs.update)(
                mean_state_pred=mean_state_pred,
                var_state_pred=var_state_pred,
                x_meas=x_meas,
                mean_meas=mean_meas,
                wgt_meas=W_meas,
                var_meas=var_meas
        )
        # output
        carry = {
            "state_filt": (mean_state_next, var_state_next)
        }
        stack = {
            "state_filt": (mean_state_next, var_state_next),
            "state_pred": (mean_state_pred, var_state_pred)
        }
        return carry, stack
    # scan initial value
    scan_init = {
        "state_filt": (mean_state_init, var_state_init)
    }
    if key is not None:
        keys = jax.random.split(key, num=n_steps)
    else:
        keys = jnp.zeros(n_steps)

    filter_kwargs = {
        "t": jnp.arange(n_steps),
        "key": keys
    }
    # scan itself
    _, scan_out = jax.lax.scan(scan_fun, scan_init, filter_kwargs)
    # append initial values to front
    scan_out["state_filt"] = (
        jnp.concatenate([mean_state_init[None], scan_out["state_filt"][0]]),
        jnp.concatenate([var_state_init[None], scan_out["state_filt"][1]])
    )
    scan_out["state_pred"] = (
        jnp.concatenate([mean_state_init[None], scan_out["state_pred"][0]]),
        jnp.concatenate([var_state_init[None], scan_out["state_pred"][1]])
    )
    return scan_out


def solve_sim(key, ode_fun, ode_weight, ode_init,
              t_min, t_max, n_steps,
              interrogate,
              prior_pars,
              kalman_type="standard", **params):
    r"""
    Draw sample solution. Same arguments as :func:`~ode.solve_mv`.

    Returns:
        (ndarray(n_steps+1, n_blocks, n_bstate)): Sample solution for :math:`X_t` at times :math:`t \in [a, b]`.

    """
    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError

    prior_weight, prior_var = prior_pars
    n_block = prior_weight.shape[0]
    key, *subkeys = jax.random.split(key, num=n_steps+1)
    # subkeys = jnp.reshape(jnp.array(subkeys), newshape=(n_steps, n_block, 2))

    # forward pass
    filt_out = _solve_filter(
        key=key,
        ode_fun=ode_fun, ode_weight=ode_weight, ode_init=ode_init,
        t_min=t_min, t_max=t_max, n_steps=n_steps,
        interrogate=interrogate,
        prior_weight=prior_weight, prior_var=prior_var,
        kalman_funs=kalman_funs, **params
    )
    mean_state_pred, var_state_pred = filt_out["state_pred"]
    mean_state_filt, var_state_filt = filt_out["state_filt"]

    # backward pass
    def scan_fun(x_state_next, smooth_kwargs):
        mean_state_filt = smooth_kwargs['mean_state_filt']
        var_state_filt = smooth_kwargs['var_state_filt']
        mean_state_pred = smooth_kwargs['mean_state_pred']
        var_state_pred = smooth_kwargs['var_state_pred']
        key = smooth_kwargs['key']

        mean_state_sim, var_state_sim = jax.vmap(kalman_funs.smooth_sim)(
            x_state_next=x_state_next,
            wgt_state=prior_weight,
            mean_state_filt=mean_state_filt,
            var_state_filt=var_state_filt,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            var_state=prior_var
        )
        x_state_curr = jax.random.multivariate_normal(key, mean_state_sim, var_state_sim, method='svd')
        return x_state_curr, x_state_curr
    # initialize
    scan_init = jax.random.multivariate_normal(
        subkeys[n_steps-1],
        mean_state_filt[n_steps],
        var_state_filt[n_steps],
        method='svd')
    
    # scan arguments
    scan_kwargs = {
        'mean_state_filt': mean_state_filt[1:n_steps],
        'var_state_filt': var_state_filt[1:n_steps],
        'mean_state_pred': mean_state_pred[2:n_steps+1],
        'var_state_pred': var_state_pred[2:n_steps+1],
        'key': jnp.array(subkeys[:n_steps-1])
    }
    # Note: initial value x0 is assumed to be known, so we don't
    # sample it.  In fact, doing so would probably fail due to cholesky
    # of a zero variance matrix...
    _, scan_out = jax.lax.scan(scan_fun, scan_init, scan_kwargs, reverse=True)

    # append initial values to front and back
    x_state_smooth = jnp.concatenate(
        [ode_init[None], scan_out, scan_init[None]]
    )
    return x_state_smooth


def solve_mv(key, ode_fun, ode_weight, ode_init,
             t_min, t_max, n_steps,
             interrogate,
             prior_pars,
             kalman_type="standard", **params):
    r"""
    Mean and variance of the stochastic ODE solver.

    Args:
        key (PRNGKey): PRNG key or None.
        ode_fun (Callable): Higher order ODE function :math:`W X_t = F(X_t, t)` taking arguments :math:`X` and :math:`t`.
        ode_weight (ndarray(n_block, n_bmeas, n_bstate)): Weight matrix defining the measure prior; :math:`W`.
        ode_init (ndarray(n_block, n_bstate)): Initial value of the state variable :math:`X_t` at time :math:`t = a`.
        t_min (float): First time point of the time interval to be evaluated; :math:`a`.
        t_max (float): Last time point of the time interval to be evaluated; :math:`b`.
        n_steps (int): Number of discretization points (:math:`N`) of the time interval that is evaluated, such that discretization timestep is :math:`dt = (b-a)/N`.
        interrogate (Callable): Function defining the interrogation method.
        prior_pars (tuple): A tuple containing the weight matrix and the variance matrix defining the solution prior; :math:`Q, R`.
        kalman_type (str): Determine which type of Kalman (standard, square-root) to use.
        params (kwargs): Optional model parameters.

    Returns:
        (tuple):
        - **mean_state_smooth** (ndarray(n_steps+1, n_block, n_bstate)): Posterior mean of the solution process :math:`X_t` at times :math:`t \in [a, b]`.
        - **var_state_smooth** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Posterior variance of the solution process at times :math:`t \in [a, b]`.

    """
    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError
    
    prior_weight, prior_var = prior_pars
    n_block, n_bstate, _ = prior_weight.shape
    # forward pass
    filt_out = _solve_filter(
        key=key,
        ode_fun=ode_fun, ode_weight=ode_weight, ode_init=ode_init,
        t_min=t_min, t_max=t_max, n_steps=n_steps,
        interrogate=interrogate,
        prior_weight=prior_weight, prior_var=prior_var,
        kalman_funs=kalman_funs, **params
    )
    mean_state_pred, var_state_pred = filt_out["state_pred"]
    mean_state_filt, var_state_filt = filt_out["state_filt"]

    # backward pass
    def scan_fun(state_next, smooth_kwargs):
        mean_state_filt = smooth_kwargs['mean_state_filt']
        var_state_filt = smooth_kwargs['var_state_filt']
        mean_state_pred = smooth_kwargs['mean_state_pred']
        var_state_pred = smooth_kwargs['var_state_pred']
        mean_state_curr, var_state_curr = jax.vmap(kalman_funs.smooth_mv)(
                mean_state_next=state_next["mean"],
                var_state_next=state_next["var"],
                wgt_state=prior_weight,
                mean_state_filt=mean_state_filt,
                var_state_filt=var_state_filt,
                mean_state_pred=mean_state_pred,
                var_state_pred=var_state_pred,
                var_state=prior_var
        )
        state_curr = {
            "mean": mean_state_curr,
            "var": var_state_curr
        }
        return state_curr, state_curr
    # initialize
    scan_init = {
        "mean": mean_state_filt[n_steps],
        "var": var_state_filt[n_steps]
    }
    # scan arguments
    scan_kwargs = {
        'mean_state_filt': mean_state_filt[1:n_steps],
        'var_state_filt': var_state_filt[1:n_steps],
        'mean_state_pred': mean_state_pred[2:n_steps+1],
        'var_state_pred': var_state_pred[2:n_steps+1]
    }
    # Note: initial value x0 is assumed to be known, so no need to smooth it
    _, scan_out = jax.lax.scan(scan_fun, scan_init, scan_kwargs,
                               reverse=True)

    # append initial values to front and back
    mean_state_smooth = jnp.concatenate(
        [ode_init[None], scan_out["mean"], scan_init["mean"][None]]
    )
    var_state_smooth = jnp.concatenate(
        [jnp.zeros((n_block, n_bstate, n_bstate))[None], scan_out["var"],
         scan_init["var"][None]]
    )
    return mean_state_smooth, var_state_smooth



import jax
import jax.numpy as jnp
from rodeo.kalmantv import standard
from rodeo.kalmantv import square_root

def magi_logdens(ode_data_subset,
                 ode_expand,
                 n_active,
                 prior_pars,
                 kalman_type,
                 **params):
    """
    Log-density of MAGI approximation.

    Args:
        ode_data_subset (ndarray(n_steps+1, n_block, n_deriv-1)): Array specifying :math:`U_{0:N}`, the subset of the solution process needed to reconstruct the entire solution with `ode_expand()`.
        ode_expand (Callable): Function taking inputs `ode_data_subset` and `**params` and returning the full solution process :math:`X_{0:N}`.
        n_active (int): Number of active derivatives -- i.e., not those zero-padded -- for the solution process.
        prior_pars (tuple): A tuple containing the weight matrix and the variance matrix defining the solution prior; :math:`Q, R`.
        kalman_type (str): Determine which type of Kalman (standard, square-root) to use.
        **params (kwargs): Parameters to pass to `ode_expand`.

    Returns:
        (float): Value of the logdensity `p(ode_data_subset, Z = 0 | params, prior_pars)`.
    """
    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError
    
    # setup
    n_vars = ode_data_subset.shape[1]
    ode_state = ode_expand(ode_data_subset, **params)
    n_deriv = ode_state.shape[2]
    # construct `*_meas` parameters
    wgt_meas = jnp.eye(n_active, n_deriv)
    wgt_meas = jnp.stack([wgt_meas] * n_vars)
    mean_meas = jnp.zeros((n_active,))
    mean_meas = jnp.stack([mean_meas] * n_vars)
    var_meas = jnp.zeros((n_active, n_active))
    var_meas = jnp.stack([var_meas] * n_vars)
    
    # construct remaining `*_state` parameters
    mean_state = jnp.zeros((n_vars, n_deriv))
    wgt_state, var_state = prior_pars

    # kalman filter
    def filter_scan(carry, x_meas):
        mean_state_past, var_state_past = carry["state"]
        # kalman predict
        mean_state_pred, var_state_pred = jax.vmap(kalman_funs.predict)(
            mean_state_past=mean_state_past,
            var_state_past=var_state_past,
            mean_state=mean_state,
            wgt_state=wgt_state,
            var_state=var_state
        )
        # kalman forecast (for logdens)
        mean_state_fore, var_state_fore = jax.vmap(kalman_funs.forecast)(
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            mean_meas=mean_meas,
            wgt_meas=wgt_meas,
            var_meas=var_meas
        )
        # var_state_fore = jax.vmap(lambda a: a.dot(a.T))(var_state_fore)
        logdens = jax.vmap(jax.scipy.stats.multivariate_normal.logpdf)(
            x=x_meas,
            mean=mean_state_fore,
            cov=var_state_fore
        )
        # kalman update
        mean_state_next, var_state_next = jax.vmap(kalman_funs.update)(
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            x_meas=x_meas,
            mean_meas=mean_meas,
            wgt_meas=wgt_meas,
            var_meas=var_meas
        )
        carry["state"] = (mean_state_next, var_state_next)
        carry["logdens"] = carry["logdens"] + jnp.sum(logdens)
        return carry, None

    filter_init = {
        "state": (ode_state[0], jnp.zeros((n_vars, n_deriv, n_deriv))),
        "logdens": 0.0
    }

    res, _ = jax.lax.scan(
        f=filter_scan,
        init=filter_init,
        xs=ode_state[1:, :, :n_active]
    )

    return res["logdens"]



r"""
This module combines prior parameters into a single block.
"""

import jax
import jax.scipy.linalg as jsl

def indep_init(prior_pars):
    """
    Combine blocks of prior parameters into dense matrices.

    Args:
        prior_pars (tuple): A tuple containing the weight matrix and the variance matrix defining the solution prior; :math:`Q, R`.

    Returns:
        (tuple):
        - **prior_weight** (ndarray(1, n_block * p, n_block * p)): Transition matrix defining the solution prior; :math:`Q`.
        - **prior_var** (ndarray(1, n_block * p, n_block * p)): Variance matrix defining the solution prior; :math:`R`.
    """
    prior_weight, prior_var = prior_pars
    prior_weight = jsl.block_diag(*prior_weight)[None, :]
    prior_var = jsl.block_diag(*prior_var)[None, :]
    return prior_weight, prior_var



r"""
This module implements the DALTON solver which gives an approximate likelihood of :math:`p(Y_{0:M} \mid Z_{1:N})`.

The model is

.. math::

    x_0 = v

    X_n = c_n + Q_n X_{n-1} + R_n^{1/2} \epsilon_n

    Z_n = W_n X_n - f(X_n, t_n) + V_n^{1/2} \eta_n.
    
    Y_m = g(X_m, \phi_m)

where :math:`g` is a general distribution function. In the case that :math:`g` is Gaussian, use :func:`dalton` for a better approximation. In other cases, use :func:`daltonng`. We assume that :math:`c_n = 0, Q_n = Q, R_n = R`, and :math:`W_n = W` for all :math:`n`.

In the Gaussian case, we assume the observation model is

.. math::

    Y_m = D_m X_m + \Omega^{1/2}_m \epsilon_m.

We assume that the :math:`M \leq N`, so that the observation step size is larger than that of the evaluation step size.

"""
import jax
import jax.numpy as jnp
import jax.scipy as jsp
from rodeo.kalmantv import standard
from rodeo.kalmantv import square_root
from rodeo.inference.fenrir import _forecast_update
from rodeo.utils import multivariate_normal_logpdf
from rodeo.solve import _solve_filter as _solve_filter_ode

# --- loglikelihood -----------------------------------------------------------


def dalton(key, ode_fun, ode_weight, ode_init, 
           t_min, t_max, n_steps,
           interrogate,
           prior_pars,
           obs_data, obs_times, obs_weight, obs_var,
           kalman_type="standard", **params):
    r"""
    Compute marginal loglikelihood of DALTON algorithm for Gaussian observations; :math:`p(Y_{0:M} \mid Z_{1:N})`.

    Args:
        key (PRNGKey): PRNG key.
        ode_fun (Callable): Higher order ODE function :math:`W X_t = F(X_t, t)` taking arguments :math:`X` and :math:`t`.
        ode_weight (ndarray(n_block, n_bmeas, n_bstate)): Weight matrix defining the measure prior; :math:`W`.
        ode_init (ndarray(n_block, n_bstate)): Initial value of the state variable :math:`X_t` at time :math:`t = a`.
        t_min (float): First time point of the time interval to be evaluated; :math:`a`.
        t_max (float): Last time point of the time interval to be evaluated; :math:`b`.
        n_steps (int): Number of discretization points (:math:`N`) of the time interval that is evaluated, such that discretization timestep is :math:`dt = (b-a)/N`.
        interrogate (Callable): Function defining the interrogation method.
        prior_pars (tuple): A tuple containing the weight matrix and the variance matrix defining the solution prior; :math:`Q, R`.
        obs_data (ndarray(n_obs, n_blocks, n_bobs)): Observed data; :math:`Y_{0:M}`.
        obs_times (ndarray(n_obs)): Observation time; :math:`0, \ldots, M`.
        obs_weight (ndarray(n_obs, n_blocks, n_bobs, n_bstate)): Weight matrix in the observation model; :math:`D_{0:M}`.
        obs_var (ndarry(n_obs, n_blocks, n_bobs, n_bobs)): Variance matrix in the observation model; :math:`\Omega_{0:M}`
        kalman_type (str): Determine which type of Kalman (standard, square-root) to use.
        params (kwargs): Optional model parameters.

    Returns:
        (float): Loglikelihood of :math:`p(Y_{0:M} \mid Z_{1:N})`.

    """
    # Dimensions of block, state and measure variables
    n_block, n_bmeas, n_bstate = ode_weight.shape
    # Dimension of observation
    n_bobs = obs_weight.shape[2]

    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError
    
    # prior variables
    prior_weight, prior_var = prior_pars
    
    # insert observations on solver time grid
    sim_times = jnp.linspace(t_min, t_max, n_steps + 1)
    obs_ind = jnp.searchsorted(sim_times, obs_times)

    # arguments for kalman_filter and kalman_smooth
    x_meas = jnp.zeros((n_block, n_bmeas))
    obs_mean = jnp.zeros((n_block, n_bobs))
    mean_state = jnp.zeros((n_block, n_bstate))
    mean_state_init = ode_init
    var_state_init = jnp.zeros((n_block, n_bstate, n_bstate))

    # forecast function without kalman_funs
    forecast_update = lambda mean_state_pred, var_state_pred,\
                             x_meas, mean_meas, wgt_meas, var_meas\
                             : _forecast_update(mean_state_pred, var_state_pred,
                                                x_meas, mean_meas,
                                                wgt_meas, var_meas,
                                                kalman_funs)

    # compute p(Z_{1:N}, Y_{0:M})
    def scan(carry, filter_kwargs):
        mean_state_filt_zy, var_state_filt_zy = carry["state_filt_joint"]
        mean_state_filt_z, var_state_filt_z = carry["state_filt_marg"]
        logdens_zy = carry["logdens_joint"]
        logdens_z = carry["logdens_marg"]
        t = filter_kwargs["t"]
        keys = filter_kwargs["key"]
        i = carry["i"]
        ode_time = t_min + (t_max-t_min)*(t+1)/n_steps
        
        # compute joint logpdf
        mean_state_pred_zy, var_state_pred_zy = jax.vmap(kalman_funs.predict)(
                mean_state_past=mean_state_filt_zy,
                var_state_past=var_state_filt_zy,
                mean_state=mean_state,
                wgt_state=prior_weight,
                var_state=prior_var
        )
        # compute meas parameters
        wgt_meas, mean_meas, var_meas = interrogate(
            key=keys[0],
            ode_fun=ode_fun,
            ode_weight=ode_weight,
            t=ode_time,
            mean_state_pred=mean_state_pred_zy,
            var_state_pred=var_state_pred_zy,
            **params
        )
        W_meas = ode_weight + wgt_meas

        # both z and y are observed
        def zy_update():
            wgt_meas_obs = jnp.concatenate([W_meas, obs_weight[i]], axis=1)
            mean_meas_obs = jnp.concatenate([mean_meas, obs_mean], axis=1)
            var_meas_obs = jax.vmap(jsp.linalg.block_diag)(var_meas, obs_var[i])
            x_meas_obs = jnp.concatenate([x_meas, obs_data[i]], axis=1)
            logp, mean_state_next, var_state_next = jax.vmap(forecast_update)(
                    mean_state_pred=mean_state_pred_zy,
                    var_state_pred=var_state_pred_zy,
                    x_meas=x_meas_obs,
                    mean_meas=mean_meas_obs,
                    wgt_meas=wgt_meas_obs,
                    var_meas=var_meas_obs
            )                     
            return mean_state_next, var_state_next, jnp.sum(logp), i+1

        # only z is observed
        def z_update():
            logp, mean_state_next, var_state_next = jax.vmap(forecast_update)(
                    mean_state_pred=mean_state_pred_zy,
                    var_state_pred=var_state_pred_zy,
                    x_meas=x_meas,
                    mean_meas=mean_meas,
                    wgt_meas=W_meas,
                    var_meas=var_meas                                               
            )
            return mean_state_next, var_state_next, jnp.sum(logp), i

        mean_state_next_zy, var_state_next_zy, logp, i = jax.lax.cond(t+1 == obs_ind[i], zy_update, z_update)
        logdens_zy += logp
        

        # compute marginal logpdf
        mean_state_pred_z, var_state_pred_z = jax.vmap(kalman_funs.predict)(
                mean_state_past=mean_state_filt_z,
                var_state_past=var_state_filt_z,
                mean_state=mean_state,
                wgt_state=prior_weight,
                var_state=prior_var
        )
        # compute meas parameters
        wgt_meas, mean_meas, var_meas = interrogate(
            key=keys[1],
            ode_fun=ode_fun,
            ode_weight=ode_weight,
            t=ode_time,
            mean_state_pred=mean_state_pred_z,
            var_state_pred=var_state_pred_z,
            **params
        )
        W_meas = ode_weight + wgt_meas
        # kalman forecast and update
        logp, mean_state_next_z, var_state_next_z = jax.vmap(forecast_update)(
                mean_state_pred=mean_state_pred_z,
                var_state_pred=var_state_pred_z,
                x_meas=x_meas,
                mean_meas=mean_meas,
                wgt_meas=W_meas,
                var_meas=var_meas
        )                       
        logdens_z += jnp.sum(logp)
        # carry over state
        carry = {
            "state_filt_joint": (mean_state_next_zy, var_state_next_zy),
            "state_filt_marg": (mean_state_next_z, var_state_next_z),
            "logdens_joint": logdens_zy,
            "logdens_marg": logdens_z,
            "i": i
        }
        return carry, None
    
    # compute log-density of p(Y_0 |X_0) if Y_0 is at time 0
    def _logy0():
        logdens_zy = jnp.sum(
            jax.vmap(lambda b:
                    multivariate_normal_logpdf(obs_data[0, b], mean=obs_weight[0, b].dot(ode_init[b]) + obs_mean[b], cov=obs_var[0, b])
            )(jnp.arange(n_block)))
        return logdens_zy, 1
    def _no_logy0():
        return 0.0, 0
    logdens_zy, i = jax.lax.cond(obs_ind[0] == 0, _logy0, _no_logy0)
    
    scan_init = {
        "state_filt_joint": (mean_state_init, var_state_init),
        "state_filt_marg": (mean_state_init, var_state_init),
        "logdens_joint": logdens_zy,
        "logdens_marg": 0.0,
        "i": i
    }

    if key is not None:
        keys = jax.random.split(key, num=(n_steps, 2))
    else:
        keys = jnp.zeros((n_steps, 2))

    filter_kwargs = {
        "t": jnp.arange(n_steps),
        "key": keys
    }
    out, _ = jax.lax.scan(scan, scan_init, filter_kwargs)
    return out["logdens_joint"] - out["logdens_marg"]


# --- ODE solver --------------------------------------------------------------


# use linearizations and observations
def _solve_filter(key, ode_fun, ode_weight, ode_init, 
                  t_min, t_max, n_steps,
                  interrogate,
                  prior_weight, prior_var,
                  obs_data, obs_times, obs_weight, obs_var,
                  kalman_funs, **params):
    r"""
    Forward pass of the DALTON algorithm with Gaussian observations. Same arguments as :func:`dalton`.

    Returns:
        (tuple):
        - **mean_state_pred** (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time t given observations from times [a...t-1] for :math:`t \in [a, b]`.
        - **var_state_pred** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Variance estimate for state at time t given observations from times [a...t-1] for :math:`t \in [a, b]`.
        - **mean_state_filt** (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time t given observations from times [a...t] for :math:`t \in [a, b]`.
        - **var_state_filt** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Variance estimate for state at time t given observations from times [a...t] for :math:`t \in [a, b]`.

    """
    # Dimensions of block, state and measure variables
    n_block, n_bmeas, n_bstate = ode_weight.shape
    # Dimension of observation
    n_bobs = obs_weight.shape[2]

    # insert observations on solver time grid
    sim_times = jnp.linspace(t_min, t_max, n_steps + 1)
    obs_ind = jnp.searchsorted(sim_times, obs_times)

    # arguments for kalman_filter and kalman_smooth
    x_meas = jnp.zeros((n_block, n_bmeas))
    obs_mean = jnp.zeros((n_block, n_bobs))
    mean_state = jnp.zeros((n_block, n_bstate))
    mean_state_init = ode_init
    var_state_init = jnp.zeros((n_block, n_bstate, n_bstate))

    # compute p(X_{1:n} | Z_{1:n}, Y_{0:m})
    def scan_fun(carry, filter_kwargs):
        mean_state_filt, var_state_filt = carry["state_filt"]
        i = carry["i"]
        t = filter_kwargs["t"]
        key = filter_kwargs["key"]
        ode_time = t_min + (t_max-t_min)*(t+1)/n_steps
        
        # kalman predict
        mean_state_pred, var_state_pred = jax.vmap(kalman_funs.predict)(
                mean_state_past=mean_state_filt,
                var_state_past=var_state_filt,
                mean_state=mean_state,
                wgt_state=prior_weight,
                var_state=prior_var
        )
        # compute meas parameters
        wgt_meas, mean_meas, var_meas = interrogate(
            key=key,
            ode_fun=ode_fun,
            ode_weight=ode_weight,
            t=ode_time,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            **params
        )
        W_meas = ode_weight + wgt_meas

        # both z and y are observed
        def zy_update():
            wgt_meas_obs = jnp.concatenate([W_meas, obs_weight[i]], axis=1)
            mean_meas_obs = jnp.concatenate([mean_meas, obs_mean], axis=1)
            var_meas_obs = jax.vmap(jsp.linalg.block_diag)(var_meas, obs_var[i])
            x_meas_obs = jnp.concatenate([x_meas, obs_data[i]], axis=1)
            mean_state_next, var_state_next = jax.vmap(kalman_funs.update)(
                    mean_state_pred=mean_state_pred,
                    var_state_pred=var_state_pred,
                    x_meas=x_meas_obs,
                    mean_meas=mean_meas_obs,
                    wgt_meas=wgt_meas_obs,
                    var_meas=var_meas_obs
            )                        
            return mean_state_next, var_state_next, i+1

        # only z is observed
        def z_update():
            mean_state_next, var_state_next = jax.vmap(kalman_funs.update)(
                    mean_state_pred=mean_state_pred,
                    var_state_pred=var_state_pred,
                    x_meas=x_meas,
                    mean_meas=mean_meas,
                    wgt_meas=W_meas,
                    var_meas=var_meas
            )                            
            return mean_state_next, var_state_next, i

        mean_state_next, var_state_next, i = jax.lax.cond(t+1 == obs_ind[i], zy_update, z_update)
        # output
        carry = {
            "state_filt": (mean_state_next, var_state_next),
            "i": i
        }
        stack = {
            "state_filt": (mean_state_next, var_state_next),
            "state_pred": (mean_state_pred, var_state_pred)
        }
        return carry, stack

    # check if observations start at 0
    i = jax.lax.cond(obs_ind[0] == 0, lambda: 1, lambda: 0)

    # scan initial value for computing p(X_{0:n} | Y_{0:m}, Z_{1:n})
    scan_init = {
        "state_filt": (mean_state_init, var_state_init),
        "i": i
    }
    
    if key is not None:
        keys = jax.random.split(key, num=n_steps)
    else:
        keys = jnp.zeros(n_steps)
    
    filter_kwargs = {
        "t": jnp.arange(n_steps),
        "key": keys
    }
    _, scan_out = jax.lax.scan(scan_fun, scan_init, filter_kwargs)
    # append initial values to front
    scan_out["state_filt"] = (
        jnp.concatenate([mean_state_init[None], scan_out["state_filt"][0]]),
        jnp.concatenate([var_state_init[None], scan_out["state_filt"][1]])
    )
    scan_out["state_pred"] = (
        jnp.concatenate([mean_state_init[None], scan_out["state_pred"][0]]),
        jnp.concatenate([var_state_init[None], scan_out["state_pred"][1]])
    )
    return scan_out


def solve_mv(key, ode_fun, ode_weight, ode_init, 
             t_min, t_max, n_steps,
             interrogate,
             prior_pars,
             obs_data, obs_times, obs_weight, obs_var,
             kalman_type="standard", **params):
    r"""
    DALTON algorithm to compute the mean and variance of :math:`p(X_{0:N} \mid Y_{0:M}, Z_{1:N})` assuming Gaussian observations.
    Same arguments as :func:`dalton`.

    Returns:
        (tuple):
        - **mean_state_smooth** (ndarray(n_steps+1, n_block, n_bstate)): Posterior mean of the solution process :math:`X_t` at times :math:`t \in [a, b]`.
        - **var_state_smooth** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Posterior variance of the solution process at times :math:`t \in [a, b]`.

    """
    
    prior_weight, prior_var = prior_pars
    n_block, n_bstate, _ = prior_weight.shape

    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError

    # forward pass
    filt_out = _solve_filter(
        key=key,
        ode_fun=ode_fun, ode_weight=ode_weight, ode_init=ode_init,
        t_min=t_min, t_max=t_max, n_steps=n_steps, 
        interrogate=interrogate,
        prior_weight=prior_weight, prior_var=prior_var,
        obs_data=obs_data, obs_times=obs_times,
        obs_weight=obs_weight, obs_var=obs_var,
        kalman_funs=kalman_funs, **params   
    )
    mean_state_pred, var_state_pred = filt_out["state_pred"]
    mean_state_filt, var_state_filt = filt_out["state_filt"]

    # backward pass
    def scan_fun(state_next, smooth_kwargs):
        mean_state_filt = smooth_kwargs['mean_state_filt']
        var_state_filt = smooth_kwargs['var_state_filt']
        mean_state_pred = smooth_kwargs['mean_state_pred']
        var_state_pred = smooth_kwargs['var_state_pred']
        mean_state_curr, var_state_curr = jax.vmap(kalman_funs.smooth_mv)(
                mean_state_next=state_next["mean"],
                var_state_next=state_next["var"],
                wgt_state=prior_weight,
                mean_state_filt=mean_state_filt,
                var_state_filt=var_state_filt,
                mean_state_pred=mean_state_pred,
                var_state_pred=var_state_pred,
                var_State=prior_var
        )
        state_curr = {
            "mean": mean_state_curr,
            "var": var_state_curr
        }
        return state_curr, state_curr
    # initialize
    scan_init = {
        "mean": mean_state_filt[n_steps],
        "var": var_state_filt[n_steps]
    }
    # scan arguments
    scan_kwargs = {
        'mean_state_filt': mean_state_filt[1:n_steps],
        'var_state_filt': var_state_filt[1:n_steps],
        'mean_state_pred': mean_state_pred[2:n_steps+1],
        'var_state_pred': var_state_pred[2:n_steps+1]
    }
    # Note: initial value x0 is assumed to be known, so no need to smooth it
    _, scan_out = jax.lax.scan(scan_fun, scan_init, scan_kwargs, reverse=True)

    # append initial values to front and back
    mean_state_smooth = jnp.concatenate(
        [ode_init[None], scan_out["mean"], scan_init["mean"][None]]
    )
    var_state_smooth = jnp.concatenate(
        [jnp.zeros((n_block, n_bstate, n_bstate))[None], scan_out["var"],
         scan_init["var"][None]]
    )
    return mean_state_smooth, var_state_smooth


def solve_sim(key, ode_fun, ode_weight, ode_init, 
              t_min, t_max, n_steps,
              interrogate,
              prior_pars,
              obs_data, obs_times, obs_weight, obs_var,
              kalman_type="standard", **params):
    r"""
    DALTON algorithm to sample from :math:`p(X_{0:N} \mid Y_{0:M}, Z_{1:N})` assuming Gaussian observations.
    Same arguments as :func:`dalton`.

    Returns:
        (tuple):
        - **mean_state_smooth** (ndarray(n_steps+1, n_block, n_bstate)): Posterior mean of the solution process :math:`X_t` at times :math:`t \in [a, b]`.
        - **var_state_smooth** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Posterior variance of the solution process at times :math:`t \in [a, b]`.

    """
    prior_weight, prior_var = prior_pars
    
    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError
    
    key, *subkeys = jax.random.split(key, num=n_steps+1)
    # forward pass
    filt_out = _solve_filter(
        key=key,
        ode_fun=ode_fun, ode_weight=ode_weight, ode_init=ode_init,
        t_min=t_min, t_max=t_max, n_steps=n_steps, 
        interrogate=interrogate,
        prior_weight=prior_weight, prior_var=prior_var,
        obs_data=obs_data, obs_times=obs_times,
        obs_weight=obs_weight, obs_var=obs_var,
        kalman_funs=kalman_funs, **params   
    )
    mean_state_pred, var_state_pred = filt_out["state_pred"]
    mean_state_filt, var_state_filt = filt_out["state_filt"]

    # backward pass
    def scan_fun(x_state_next, smooth_kwargs):
        mean_state_filt = smooth_kwargs['mean_state_filt']
        var_state_filt = smooth_kwargs['var_state_filt']
        mean_state_pred = smooth_kwargs['mean_state_pred']
        var_state_pred = smooth_kwargs['var_state_pred']

        key = smooth_kwargs['key']

        mean_state_sim, var_state_sim = jax.vmap(kalman_funs.smooth_sim)(
            x_state_next=x_state_next,
            wgt_state=prior_weight,
            mean_state_filt=mean_state_filt,
            var_state_filt=var_state_filt,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            var_state=prior_var
        )
        x_state_curr = jax.random.multivariate_normal(key, mean_state_sim, var_state_sim, method='svd')
        return x_state_curr, x_state_curr
    # initialize
    scan_init = jax.random.multivariate_normal(
        subkeys[n_steps-1],
        mean_state_filt[n_steps],
        var_state_filt[n_steps],
        method='svd')
    # scan arguments
    scan_kwargs = {
        'mean_state_filt': mean_state_filt[1:n_steps],
        'var_state_filt': var_state_filt[1:n_steps],
        'mean_state_pred': mean_state_pred[2:n_steps+1],
        'var_state_pred': var_state_pred[2:n_steps+1],
        'key': jnp.array(subkeys[:n_steps-1])
    }
    # Note: initial value x0 is assumed to be known, so no need to smooth it
    _, scan_out = jax.lax.scan(scan_fun, scan_init, scan_kwargs, reverse=True)

    # append initial values to front and back
    x_state_smooth = jnp.concatenate(
        [ode_init[None], scan_out, scan_init[None]]
    )
    return x_state_smooth

# --- non-Gaussian loglikelihood -------------------------------------------------


def _solve_filter_nn(key, ode_fun, ode_weight, ode_init, 
                     t_min, t_max, n_steps,
                     interrogate,
                     prior_weight, prior_var,
                     obs_data, obs_times, obs_loglik_i,
                     kalman_funs, **params):
    r"""
    Forward pass of the DALTON algorithm using non-Gaussian observations. Same arguments as :func:`daltonng`.

    Returns:
        (tuple):
        - **mean_state_pred** (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time t given observations from times [a...t-1] for :math:`t \in [a, b]`.
        - **var_state_pred** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Variance estimate for state at time t given observations from times [a...t-1] for :math:`t \in [a, b]`.
        - **mean_state_filt** (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time t given observations from times [a...t] for :math:`t \in [a, b]`.
        - **var_state_filt** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Variance estimate for state at time t given observations from times [a...t] for :math:`t \in [a, b]`.

    """
    # Dimensions of block, state and measure variables
    n_block, n_bmeas, n_bstate = ode_weight.shape
    # Dimension of observation
    # n_bobs = obs_data.shape[2]

    # insert observations on solver time grid
    sim_times = jnp.linspace(t_min, t_max, n_steps + 1)
    obs_ind = jnp.searchsorted(sim_times, obs_times)

    # arguments for kalman_filter and kalman_smooth
    x_meas = jnp.zeros((n_block, n_bmeas))
    obs_mean = jnp.zeros((n_block, n_bstate))
    mean_state = jnp.zeros((n_block, n_bstate))
    mean_state_init = ode_init
    var_state_init = jnp.zeros((n_block, n_bstate, n_bstate))

    # compute p(X_{1:n} | Z_{1:n}, \hat Y_{0:m})
    def scan_fun(carry, filter_kwargs):
        mean_state_filt, var_state_filt = carry["state_filt"]
        i = carry["i"]
        t = filter_kwargs["t"]
        key = filter_kwargs["key"]
        ode_time = t_min + (t_max-t_min)*(t+1)/n_steps
        
        # kalman predict
        mean_state_pred, var_state_pred = jax.vmap(kalman_funs.predict)(
            mean_state_past=mean_state_filt,
            var_state_past=var_state_filt,
            mean_state=mean_state,
            wgt_state=prior_weight,
            var_state=prior_var
        )
        # compute meas parameters
        wgt_meas, mean_meas, var_meas = interrogate(
            key=key,
            ode_fun=ode_fun,
            ode_weight=ode_weight,
            t=ode_time,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            **params
        )
        W_meas = ode_weight + wgt_meas

        # both z and y are observed
        def zy_update():
            # transform to yhat
            obs_grad = jax.jacrev(obs_loglik_i, argnums=1)(obs_data[i], mean_state_pred, i, **params)
            obs_hes = jax.jacfwd(jax.jacrev(obs_loglik_i, argnums=1), argnums=1)(
                obs_data[i], mean_state_pred, i, **params
            )
            obs_var = jax.vmap(lambda b: -jnp.linalg.pinv(obs_hes[b, :, b]))(jnp.arange(n_block))
            obs_weight = jnp.where(obs_var != 0, 1, 0)
            obs_hat = jax.vmap(lambda b: 
                obs_weight[i].dot(mean_state_pred[b]) + obs_var[b].dot(obs_grad[b])
            )(jnp.arange(n_block))
            
            # Cmu = jax.vmap(lambda b: wgt_curr[b].dot(mean_state_pred[b]))(jnp.arange(n_block))
            # gpmu = jax.jacfwd(fun_obs)(Cmu, y_curr, theta, i)
            # gppmu = jax.jacfwd(jax.jacrev(fun_obs))(Cmu, y_curr, theta, i)
            # var_obs = jax.vmap(lambda b: -jnp.linalg.pinv(gppmu[b, :, b]))(jnp.arange(n_block))
            # y_new = jax.vmap(lambda b: Cmu[b] + var_obs[b].dot(gpmu[b]))(jnp.arange(n_block))
            # stack measure and observation variables
            wgt_meas_obs = jnp.concatenate([W_meas, obs_weight], axis=1)
            mean_meas_obs = jnp.concatenate([mean_meas, obs_mean], axis=1)
            var_meas_obs = jax.vmap(lambda b: jsp.linalg.block_diag(var_meas[b], obs_var[b]))(jnp.arange(n_block))
            x_meas_obs = jnp.concatenate([x_meas, obs_hat], axis=1)
            # kalman update
            mean_state_next, var_state_next = jax.vmap(kalman_funs.update)(
                mean_state_pred=mean_state_pred,
                var_state_pred=var_state_pred,
                x_meas=x_meas_obs,
                mean_meas=mean_meas_obs,
                wgt_meas=wgt_meas_obs,
                var_meas=var_meas_obs
            )
            return mean_state_next, var_state_next, i+1
        
        # only z is observed
        def z_update():
            # kalman update
            mean_state_next, var_state_next = jax.vmap(kalman_funs.update)(
                mean_state_pred=mean_state_pred,
                var_state_pred=var_state_pred,
                x_meas=x_meas,
                mean_meas=mean_meas,
                wgt_meas=W_meas,
                var_meas=var_meas
            )
            return mean_state_next, var_state_next, i

        mean_state_next, var_state_next, i = jax.lax.cond(t+1 == obs_ind[i], zy_update, z_update)
        # output
        carry = {
            "state_filt": (mean_state_next, var_state_next),
            "i": i
        }
        stack = {
            "state_filt": (mean_state_next, var_state_next),
            "state_pred": (mean_state_pred, var_state_pred)
        }
        return carry, stack

    # check if observations start at 0
    i = jax.lax.cond(obs_ind[0] == 0, lambda: 1, lambda: 0)

    # scan initial value for computing p(X_{0:n} | \hat Y_{0:m}, Z_{1:n})
    scan_init = {
        "state_filt": (mean_state_init, var_state_init),
        "i": i
    }
    
    if key is not None:
        keys = jax.random.split(key, num=n_steps)
    else:
        keys = jnp.zeros(n_steps)
    
    filter_kwargs = {
        "t": jnp.arange(n_steps),
        "key": keys
    }
    _, scan_out = jax.lax.scan(scan_fun, scan_init, filter_kwargs)
    # append initial values to front
    scan_out["state_filt"] = (
        jnp.concatenate([mean_state_init[None], scan_out["state_filt"][0]]),
        jnp.concatenate([var_state_init[None], scan_out["state_filt"][1]])
    )
    scan_out["state_pred"] = (
        jnp.concatenate([mean_state_init[None], scan_out["state_pred"][0]]),
        jnp.concatenate([var_state_init[None], scan_out["state_pred"][1]])
    )
    return scan_out


def _logx_yhat(mean_state_filt, var_state_filt,
               mean_state_pred, var_state_pred,
               prior_weight, prior_var, kalman_funs):
    r"""
    Compute the loglikelihood of :math:`p(X_{0:N} \mid \hat Y_{0:M}, Z_{1:N})`.
    
    Args:
        mean_state_pred (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time n given observations from times [0...n]; denoted by :math:`\mu_{n|n-1}`.
        var_state_pred (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Covariance of estimate for state at time n given observations from times [0...n-1]; denoted by :math:`\Sigma_{n|n-1}`.
        mean_state_filt (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time n given observations from times [0...n]; denoted by :math:`\mu_{n|n}`.
        var_state_filt (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Covariance of estimate for state at time n given observations from times [0...n]; denoted by :math:`\Sigma_{n|n}`.
        prior_weight (ndarray(n_block, n_bstate, n_bstate)): Weight matrix defining the solution prior; :math:`Q`.
        prior_var (ndarray(n_block, n_bstate, n_bstate)): Variance matrix defining the solution prior; :math:`R`.
        kalman_funs (object): An object that contains the Kalman filtering functions: predict, update and smooth.

    Returns:
        (tuple):
        - **mean_state_smooth** (ndarray(n_steps+1, n_block, n_bstate)): Posterior mean of the solution process :math:`p(X_{0:N} \mid \hat Y_{0:M}, Z_{1:N})`.
        - **logx_yhat** (float): Loglikelihood of :math:`p(X_{0:N} \mid \hat Y_{0:M}, Z_{1:N})`.
    """
    # dimensions
    n_tot, n_block, _ = mean_state_filt.shape
    n_steps = n_tot - 1

    # backward pass
    def scan_fun(state_next, smooth_kwargs):
        mean_state_filt = smooth_kwargs['mean_state_filt']
        var_state_filt = smooth_kwargs['var_state_filt']
        mean_state_pred = smooth_kwargs['mean_state_pred']
        var_state_pred = smooth_kwargs['var_state_pred']
        logx_yhat = state_next["logx_yhat"]
        mean_state_curr, var_state_curr = jax.vmap(kalman_funs.smooth_mv)(
            mean_state_next=state_next["mean"],
            var_state_next=state_next["var"],
            mean_state_filt=mean_state_filt,
            var_state_filt=var_state_filt,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            wgt_state=prior_weight,
            var_state=prior_var
        )
        mean_state_sim, var_state_sim = jax.vmap(kalman_funs.smooth_sim)(
            x_state_next=state_next["mean"],
            mean_state_filt=mean_state_filt,
            var_state_filt=var_state_filt,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            wgt_state=prior_weight,
            var_state=prior_var
        )
        logx_yhat += jnp.sum(
            jax.vmap(multivariate_normal_logpdf)(mean_state_curr, mean=mean_state_sim, cov=var_state_sim)
        )
        carry = {
            "mean": mean_state_curr,
            "var": var_state_curr,
            "logx_yhat":  logx_yhat
        }
        return carry, carry
    # compute log(mu_{N|N}) at the last filtering step
    logx_yhatN = jnp.sum(
        jax.vmap(multivariate_normal_logpdf)(mean_state_filt[n_steps], mean=mean_state_filt[n_steps], cov=var_state_filt[n_steps])
    )
    # initialize
    scan_init = {
        "mean": mean_state_filt[n_steps],
        "var": var_state_filt[n_steps],
        "logx_yhat": logx_yhatN
    }
    # scan arguments
    scan_kwargs = {
        'mean_state_filt': mean_state_filt[1:n_steps],
        'var_state_filt': var_state_filt[1:n_steps],
        'mean_state_pred': mean_state_pred[2:n_steps+1],
        'var_state_pred': var_state_pred[2:n_steps+1]
    }
    # Note: initial value x0 is assumed to be known, so no need to smooth it
    last_scan, scan_out = jax.lax.scan(scan_fun, scan_init, scan_kwargs,reverse=True)

    # append initial values to front and terminal value to the back
    scan_out["mean"] = jnp.concatenate(
        [mean_state_filt[0][None], scan_out["mean"], scan_init["mean"][None]]
    )
    return scan_out["mean"], last_scan["logx_yhat"]


def _logx_z(uncond_mean, 
            mean_state_filt, var_state_filt,
            mean_state_pred, var_state_pred,
            prior_weight, prior_var, kalman_funs):
    r"""
    Compute the loglikelihood of :math:`p(X_{0:N} \mid Z_{1:N})`.
    
    Args:
        uncond_mean (ndarray(n_steps+1, n_block, n_bstate)): Unconditional mean computed from :math:`p(X_{0:N} \mid \hat Y_{0:M}, Z_{1:N})`.
        mean_state_pred (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time n given observations from times [0...n]; denoted by :math:`\mu_{n|n-1}`.
        var_state_pred (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Covariance of estimate for state at time n given observations from times [0...n-1]; denoted by :math:`\Sigma_{n|n-1}`.
        mean_state_filt (ndarray(n_steps+1, n_block, n_bstate)): Mean estimate for state at time n given observations from times [0...n]; denoted by :math:`\mu_{n|n}`.
        var_state_filt (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Covariance of estimate for state at time n given observations from times [0...n]; denoted by :math:`\Sigma_{n|n}`.
        prior_weight (ndarray(n_block, n_bstate, n_bstate)): Weight matrix defining the solution prior; :math:`Q`.
        prior_var (ndarray(n_block, n_bstate, n_bstate)): Variance matrix defining the solution prior; :math:`R`.
        kalman_funs (object): An object that contains the Kalman filtering functions: predict, update and smooth.
    
    Return:
        (float): Loglikelihood of :math:`p(X_{0:N} \mid Z_{1:N})`.

    """
    # dimensions
    n_tot, n_block, _ = mean_state_filt.shape
    n_steps = n_tot - 1

    # backward pass
    def scan_fun(logx_z, smooth_kwargs):
        mean_state_filt = smooth_kwargs['mean_state_filt']
        var_state_filt = smooth_kwargs['var_state_filt']
        mean_state_pred = smooth_kwargs['mean_state_pred']
        var_state_pred = smooth_kwargs['var_state_pred']
        uncond_next = smooth_kwargs['uncond_next']
        uncond_curr = smooth_kwargs['uncond_curr']
        mean_state_sim, var_state_sim = jax.vmap(kalman_funs.smooth_sim)(
            x_state_next=uncond_next,
            mean_state_filt=mean_state_filt,
            var_state_filt=var_state_filt,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            wgt_state=prior_weight,
            var_state=prior_var
        )
        logx_z += jnp.sum(
            jax.vmap(multivariate_normal_logpdf)(uncond_curr, mean=mean_state_sim, cov=var_state_sim)
        )
        return logx_z, logx_z
    # compute log(mu_{N|N}) at the last filtering step
    logx_zN = jnp.sum(
        jax.vmap(multivariate_normal_logpdf)(uncond_mean[n_steps], mean=mean_state_filt[n_steps], cov=var_state_filt[n_steps])
    )
    # scan arguments
    scan_kwargs = {
        'mean_state_filt': mean_state_filt[1:n_steps],
        'var_state_filt': var_state_filt[1:n_steps],
        'mean_state_pred': mean_state_pred[2:n_steps+1],
        'var_state_pred': var_state_pred[2:n_steps+1],
        'uncond_next': uncond_mean[2:n_steps+1],
        'uncond_curr': uncond_mean[1:n_steps]
    }
    # Note: initial value x0 is assumed to be known, so no need to smooth it
    scan_out, _ = jax.lax.scan(scan_fun, logx_zN, scan_kwargs, reverse=True)
    
    return scan_out

def daltonng(key, ode_fun, ode_weight, ode_init, 
             t_min, t_max, n_steps,
             interrogate,
             prior_pars,
             obs_data, obs_times, obs_loglik_i,
             kalman_type="standard", **params):
    r"""
    Compute marginal loglikelihood of DALTON algorithm for non-Gaussian observations; :math:`p(\hat Y_{0:M} \mid Z_{1:N})`.

    Args:
        key (PRNGKey): PRNG key.
        ode_fun (Callable): Higher order ODE function :math:`W X_t = F(X_t, t)` taking arguments :math:`X` and :math:`t`.
        ode_weight (ndarray(n_block, n_bmeas, n_bstate)): Weight matrix defining the measure prior; :math:`W`.
        ode_init (ndarray(n_block, n_bstate)): Initial value of the state variable :math:`X_t` at time :math:`t = a`.
        t_min (float): First time point of the time interval to be evaluated; :math:`a`.
        t_max (float): Last time point of the time interval to be evaluated; :math:`b`.
        n_steps (int): Number of discretization points (:math:`N`) of the time interval that is evaluated, such that discretization timestep is :math:`dt = (b-a)/N`.
        interrogate (Callable): Function defining the interrogation method.
        prior_pars (tuple): A tuple containing the weight matrix and the variance matrix defining the solution prior; :math:`Q, R`.
        obs_data (ndarray(n_obs, n_blocks, n_bobs)): Observed data; :math:`Y_{0:M}`.
        obs_times (ndarray(n_obs)): Observation time; :math:`0, \ldots, M`.
        obs_loglik_i (Callable): Loglikelihood function for each observation.
        kalman_type (str): Determine which type of Kalman (standard, square-root) to use.
        params (kwargs): Optional model parameters.

    
    Returns:
        (float): Loglikelihood of :math:`p(\hat Y_{0:M} \mid Z_{1:N})`.

    """
    n_obs = obs_data.shape[0]

    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError

    prior_weight, prior_var = prior_pars

    # forward pass
    filt_out = _solve_filter_nn(
        key=key,
        ode_fun=ode_fun, ode_weight=ode_weight, ode_init=ode_init,
        t_min=t_min, t_max=t_max, n_steps=n_steps, 
        interrogate=interrogate,
        prior_weight=prior_weight, prior_var=prior_var,
        obs_data=obs_data, obs_times=obs_times,
        obs_loglik_i=obs_loglik_i,
        kalman_funs=kalman_funs, **params   
    )
    mean_state_pred, var_state_pred = filt_out["state_pred"]
    mean_state_filt, var_state_filt = filt_out["state_filt"]

    # logp(X_{0:N} | \hat Y_{0:M}, Z_{1:N})
    mean_state_smooth, logx_yhat = _logx_yhat(
        mean_state_filt=mean_state_filt,
        var_state_filt=var_state_filt,
        mean_state_pred=mean_state_pred,
        var_state_pred=var_state_pred,
        prior_weight=prior_weight,
        prior_var=prior_var,
        kalman_funs=kalman_funs
    )

    # logp(Y_{0:M} | X_{0:M})
    sim_times = jnp.linspace(t_min, t_max, n_steps+1)
    obs_ind = jnp.searchsorted(sim_times, obs_times)
    def vmap_fun(i):
        return obs_loglik_i(obs_data[i], mean_state_smooth[obs_ind[i]], i, **params)
    logy_x = jnp.sum(jax.vmap(vmap_fun)(jnp.arange(n_obs)))

    # logp(X_{0:N} | Z_{1:N})
    # first do forward pass without obs
    filt_out = _solve_filter_ode(
        key=key,
        ode_fun=ode_fun, ode_weight=ode_weight, ode_init=ode_init,
        t_min=t_min, t_max=t_max, n_steps=n_steps, 
        interrogate=interrogate,
        prior_weight=prior_weight, prior_var=prior_var,
        kalman_funs=kalman_funs, **params   
    )
    mean_state_pred, var_state_pred = filt_out["state_pred"]
    mean_state_filt, var_state_filt = filt_out["state_filt"]

    logx_z = _logx_z(
        uncond_mean=mean_state_smooth,
        mean_state_filt=mean_state_filt,
        var_state_filt=var_state_filt,
        mean_state_pred=mean_state_pred,
        var_state_pred=var_state_pred,
        prior_weight=prior_weight,
        prior_var=prior_var,
        kalman_funs=kalman_funs
    )

    return logy_x + logx_z - logx_yhat


# --- non-Gaussian ODE solver -------------------------------------------------


def solve_mv_nn(key, ode_fun, ode_weight, ode_init, 
                t_min, t_max, n_steps,
                interrogate,
                prior_pars,
                obs_data, obs_times, obs_loglik_i,
                kalman_type="standard", **params):
    r"""
    DALTON algorithm to compute the mean and variance of :math:`p(X_{0:N} \mid \hat Y_{0:M}, Z_{1:N})` assuming non-Gaussian observations. 
    Same arguments as :func:`daltonng`.

    Returns:
        (tuple):
        - **mean_state_smooth** (ndarray(n_steps+1, n_block, n_bstate)): Posterior mean of the solution process :math:`X_t` at times :math:`t \in [a, b]`.
        - **var_state_smooth** (ndarray(n_steps+1, n_block, n_bstate, n_bstate)): Posterior variance of the solution process at times :math:`t \in [a, b]`.

    """
    prior_weight, prior_var = prior_pars
    n_block, n_bstate, _ = prior_weight.shape
    
    # standard or square-root filter
    if kalman_type == "standard":
        kalman_funs = standard
    elif kalman_type == "square-root":
        kalman_funs = square_root
    else:
        raise NotImplementedError
    
    # forward pass
    filt_out = _solve_filter_nn(
        key=key,
        ode_fun=ode_fun, ode_weight=ode_weight, ode_init=ode_init,
        t_min=t_min, t_max=t_max, n_steps=n_steps, 
        interrogate=interrogate,
        prior_weight=prior_weight, prior_var=prior_var,
        obs_data=obs_data, obs_times=obs_times,
        obs_loglik_i=obs_loglik_i,
        kalman_funs=kalman_funs, **params   
    )
    mean_state_pred, var_state_pred = filt_out["state_pred"]
    mean_state_filt, var_state_filt = filt_out["state_filt"]
    # backward pass
    def scan_fun(state_next, smooth_kwargs):
        mean_state_filt = smooth_kwargs['mean_state_filt']
        var_state_filt = smooth_kwargs['var_state_filt']
        mean_state_pred = smooth_kwargs['mean_state_pred']
        var_state_pred = smooth_kwargs['var_state_pred']
        mean_state_curr, var_state_curr = jax.vmap(kalman_funs.smooth_mv)(
            mean_state_next=state_next["mean"],
            var_state_next=state_next["var"],
            wgt_state=prior_weight,
            mean_state_filt=mean_state_filt,
            var_state_filt=var_state_filt,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            var_state=prior_var
        )
        state_curr = {
            "mean": mean_state_curr,
            "var": var_state_curr
        }
        return state_curr, state_curr
    # initialize
    scan_init = {
        "mean": mean_state_filt[n_steps],
        "var": var_state_filt[n_steps]
    }
    # scan arguments
    scan_kwargs = {
        'mean_state_filt': mean_state_filt[1:n_steps],
        'var_state_filt': var_state_filt[1:n_steps],
        'mean_state_pred': mean_state_pred[2:n_steps+1],
        'var_state_pred': var_state_pred[2:n_steps+1]
    }
    # Note: initial value x0 is assumed to be known, so no need to smooth it
    _, scan_out = jax.lax.scan(scan_fun, scan_init, scan_kwargs, reverse=True)

    # append initial values to front and back
    mean_state_smooth = jnp.concatenate(
        [ode_init[None], scan_out["mean"], scan_init["mean"][None]]
    )
    var_state_smooth = jnp.concatenate(
        [jnp.zeros((n_block, n_bstate, n_bstate))[None], scan_out["var"],
         scan_init["var"][None]]
    )
    return mean_state_smooth, var_state_smooth



r"""
We implement some ready to use interrogation methods from our paper. We implement the interrogation method of Chkrebtii et al (2016), Schober et al (2019) and Kramer et al (2021). 

We also implement two other interrogation methods corresponding to Schober et al (2019) and Kramer et al (2021) where we instead use the variance of Chkrebtii et al (2016). 

"""


import jax
import jax.numpy as jnp


def interrogate_chkrebtii(key, ode_fun, ode_weight, t,
                          mean_state_pred, var_state_pred,
                          kalman_type, **params):
    r"""
    Interrogate method of Chkrebtii et al (2016); DOI: 10.1214/16-BA1017.

    Same arguments and returns as :func:`~ode.interrogate_rodeo`.

    """
    n_block, n_bstate = mean_state_pred.shape
    key, *subkeys = jax.random.split(key, num=n_block+1)
    subkeys = jnp.array(subkeys)
    if kalman_type == "standard":
        var_meas = jax.vmap(lambda wm, vsp:
                            jnp.atleast_2d(jnp.linalg.multi_dot([wm, vsp, wm.T])))(
            ode_weight, var_state_pred
        )
        x_state = jax.vmap(jax.random.multivariate_normal)(
            subkeys,
            mean_state_pred,
            var_state_pred
        )
    elif kalman_type == "square-root":
        var_meas = jax.vmap(lambda wm, vsp:
                            jnp.atleast_2d(jnp.linalg.multi_dot([wm, vsp])))(
            ode_weight, var_state_pred
        )
        random_norm = jax.vmap(jax.random.normal, in_axes=(0, None))(subkeys, (n_bstate,))
        x_state = jax.vmap(lambda b:
                           mean_state_pred[b] + var_meas[b].dot(random_norm[b]))(jnp.arange(n_block))
    else:
        raise NotImplementedError
    
    mean_meas = -ode_fun(x_state, t, **params)
    return jnp.zeros(ode_weight.shape), mean_meas, var_meas


def interrogate_schober(key, ode_fun, ode_weight, t,
                        mean_state_pred, var_state_pred,
                        **params):
    r"""
    Interrogate method of Schober et al (2019); DOI: https://doi.org/10.1007/s11222-017-9798-7.

    Same arguments and returns as :func:`~ode.interrogate_rodeo`.

    """
    n_block, n_bmeas, _ = ode_weight.shape
    var_meas = jnp.zeros((n_block, n_bmeas, n_bmeas))
    mean_meas = -ode_fun(mean_state_pred, t, **params)
    return jnp.zeros(ode_weight.shape), mean_meas, var_meas


def interrogate_kramer(key, ode_fun, ode_weight, t,
                       mean_state_pred, var_state_pred,
                       **params):
    r"""
    First order interrogate method of Kramer et al (2021); DOI: https://doi.org/10.48550/arXiv.2110.11812.
    Assumes off block diagonals are zero.
    Same arguments and returns as :func:`~ode.interrogate_rodeo`.

    """
    n_block, n_bmeas, n_bstate = ode_weight.shape
    fun_meas = -ode_fun(mean_state_pred, t, **params)
    jac = jax.jacfwd(ode_fun)(mean_state_pred, t, **params)
    # need to get the diagonal of jac
    jac = jax.vmap(lambda b:
                   jac[b, :, b])(jnp.arange(n_block))
    wgt_meas = -jac
    mean_meas = jax.vmap(lambda b:
                         fun_meas[b] + jac[b].dot(mean_state_pred[b]))(jnp.arange(n_block))
    var_meas = jnp.zeros((n_block, n_bmeas, n_bmeas))
    return wgt_meas, mean_meas, var_meas


def interrogate_rodeo(key, ode_fun, ode_weight, t,
                      mean_state_pred, var_state_pred,
                      **params):
    r"""
    Rodeo interrogation method.

    Args:
        key (PRNGKey): Jax PRNG key.
        ode_fun (Callable): Higher order ODE Callable :math:`W X_t = f(X_t, t, \theta)` taking arguments :math:`X` and :math:`t`.
        ode_weight (ndarray(n_block, n_bmeas, n_bstate)): Weight matrix.
        t (float): Time point.
        mean_state_pred (ndarray(n_block, n_bstate)): Mean estimate for state at time t given observations from times [a...t-1]; denoted by :math:`\mu_{t|t-1}`.
        var_state_pred (ndarray(n_block, n_bstate, n_bstate)): Covariance of estimate for state at time t given observations from times [a...t-1]; denoted by :math:`\Sigma_{t|t-1}`.
        params : Optional model parameters.

    Returns:
        (tuple):
        - **wgt_meas** (ndarray(n_block, n_bmeas, n_bstate)): Interrogation weight matrix.
        - **mean_meas** (ndarray(n_block, n_bmeas)): Interrogation offset.
        - **var_meas** (ndarray(n_block, n_bmeas, n_bmeas)): Interrogation variance.

    """
    n_block = mean_state_pred.shape[0]
    var_meas = jax.vmap(lambda wm, vsp:
                        jnp.atleast_2d(jnp.linalg.multi_dot([wm, vsp, wm.T])))(
        ode_weight, var_state_pred
    )
    mean_meas = -ode_fun(mean_state_pred, t, **params)
    return jnp.zeros(ode_weight.shape), mean_meas, var_meas


# def interrogate_rodeo2(key, ode_fun, ode_weight, t,
#                        mean_state_pred, var_state_pred,
#                        **params):
#     r"""
#     First order interrogate method of Kramer et al (2021); DOI: https://doi.org/10.48550/arXiv.2110.11812.
#     Assumes off block diagonals are zero.

#     Same arguments and returns as :func:`~ode.interrogate_rodeo`.

#     """
#     n_block, n_bmeas, n_bstate = ode_weight.shape
#     fun_meas = -ode_fun(mean_state_pred, t, **params)
#     jac = jax.jacfwd(ode_fun)(mean_state_pred, t, **params)
#     # need to get the diagonal of jac
#     jac = jax.vmap(lambda b:
#                    jac[b, :, b])(jnp.arange(n_block))
#     wgt_meas = -jac
#     mean_meas = jax.vmap(lambda b:
#                          fun_meas[b] + jac[b].dot(mean_state_pred[b]))(jnp.arange(n_block))
#     var_meas = jax.vmap(lambda wm, vsp:
#                         jnp.atleast_2d(jnp.linalg.multi_dot([wm, vsp, wm.T])))(
#         wgt_meas, var_state_pred
#     )
#     return wgt_meas, mean_meas, var_meas



import unittest
import jax
import jax.numpy as jnp
from rodeo.utils import add_sqrt

class TestAddSqrt(unittest.TestCase):
    """
    Test if add_sqrt() give same result as expected math result.
    """

    def test_add_sqrt(self):
        key = jax.random.PRNGKey(0)
        key, subkey = jax.random.split(key)
        sqrt_A = jax.random.normal(subkey, shape = (2,2))
        key, subkey = jax.random.split(key)
        sqrt_B = jax.random.normal(subkey, shape = (2,2))
        A = jnp.dot(sqrt_A, sqrt_A.T)
        B = jnp.dot(sqrt_B, sqrt_B.T)
        sqrt_AB = add_sqrt(sqrt_A, sqrt_B)
        self.assertTrue(jnp.allclose(A + B, jnp.dot(sqrt_AB, sqrt_AB.T)))
        
if __name__ == '__main__':
    unittest.main()
    
    



r"""
Time-varying Kalman filtering and smoothing algorithms.

The Gaussian state space model underlying the algorithms is

.. math::

    x_n = c_n + Q_n x_{n-1} + R_n^{1/2} \epsilon_n

    y_n = d_n + W_n x_n + V_n^{1/2} \eta_n,

where :math:`\epsilon_n \stackrel{\text{iid}}{\sim} \operatorname{Normal}(0, I_p)` and independently :math:`\eta_n \stackrel{\text{iid}}{\sim} \operatorname{Normal}(0, I_q)`.  At each time :math:`n`, only :math:`y_n` is observed.  The Kalman filtering and smoothing algorithms efficiently calculate quantities of the form :math:`\theta_{m|n} = (\mu_{m|n}, \Sigma_{m|n})`, where

.. math::

    \mu_{m|n} = E[x_m \mid y_{0:n}]

    \Sigma_{m|n} = \text{var}(x_m \mid y_{0:n}),

for different combinations of :math:`m` and :math:`n`.

"""
import jax
import jax.numpy as jnp
from rodeo.utils import solve_var


# --- core functions -----------------------------------------------------------


def predict(mean_state_past,
            var_state_past,
            mean_state,
            wgt_state,
            var_state,
            *args, **kwargs):
    r"""
    Performs one prediction step of the Kalman filter.

    Calculates the mean and variance of :math:`p(X_n | Z_{0:n-1})` from :math:`p(X_{n-1} | Z_{0:n-1})`.

    Args:
        mean_state_past (ndarray(n_state)): State mean at time :math:`t = n-1` given observations at times :math:`t = 0, \dots, n-1`.
        var_state_past (ndarray(n_state, n_state)): State variance at time :math:`t = n-1` given observations at times :math:`t = 0, \dots, n-1`.
        mean_state (ndarray(n_state)): State equation offset at time :math:`t = n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n`.
        var_state (ndarray(n_state, n_state)): State variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_pred** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        - **var_state_pred** (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
    """

    mean_state_pred = wgt_state.dot(mean_state_past) + mean_state
    var_state_pred = jnp.linalg.multi_dot(
        [wgt_state, var_state_past, wgt_state.T]) + var_state
    return mean_state_pred, var_state_pred


def update(mean_state_pred,
           var_state_pred,
           x_meas,
           mean_meas,
           wgt_meas,
           var_meas,
           *args, **kwargs):
    r"""
    Performs one update step of the Kalman filter.

    Calculates the mean and variance of :math:`p(X_n | Z_{0:n})` from :math:`p(X_n | Z_{0:n-1})`.

    Args:
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        var_state_pred (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        x_meas (ndarray(n_meas)): Interrogated measure vector from :math:`x_{\text{state}}` at time :math:`t = n`.
        mean_meas (ndarray(n_meas)): Measurement equation offset at time :math:`t = n`.
        wgt_meas (ndarray(n_meas, n_state)): Measurement transition matrix at time :math:`t = n`.
        var_meas (ndarray(n_meas, n_meas)): Measurement variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_filt** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        - **var_state_filt** (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
    """

    # wgt_meas = W + wgt_meas
    # mean_meas_pred = W.dot(mean_state_pred) + mean_meas
    mean_meas_pred = wgt_meas.dot(mean_state_pred) + mean_meas
    var_meas_state_pred = wgt_meas.dot(var_state_pred)
    var_meas_meas_pred = jnp.linalg.multi_dot(
        [wgt_meas, var_state_pred, wgt_meas.T]) + var_meas
    var_state_meas_pred = var_state_pred.dot(wgt_meas.T)
    var_state_temp = solve_var(var_meas_meas_pred, var_state_meas_pred.T).T
    mean_state_filt = mean_state_pred + \
        var_state_temp.dot(x_meas - mean_meas_pred)
    var_state_filt = var_state_pred - \
        var_state_temp.dot(var_meas_state_pred)
    return mean_state_filt, var_state_filt


def filter(mean_state_past,
           var_state_past,
           mean_state,
           wgt_state,
           var_state,
           x_meas,
           mean_meas,
           wgt_meas,
           var_meas,
           *args, **kwargs):
    r"""
    Performs one step of the Kalman filter.

    Combines the :func:`predict` and :func:`update` steps at time :math:`t = n`.

    Args:
        mean_state_past (ndarray(n_state)): State mean at time :math:`t = n-1` given observations at times :math:`t = 0, \dots, n-1`.
        var_state_past (ndarray(n_state, n_state)): State variance at time :math:`t = n-1` given observations at times :math:`t = 0, \dots, n-1`.
        mean_state (ndarray(n_state)): State equation offset at time :math:`t = n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n`.
        var_state (ndarray(n_state, n_state)): State variance matrix at time :math:`t = n`.
        x_meas (ndarray(n_state)): Interrogated measure vector from :math:`x_{\text{state}}` at time :math:`t = n`.
        mean_meas (ndarray(n_state)): Transition offsets defining the measure prior at time :math:`t = n`.
        wgt_meas (ndarray(n_meas, n_state)): Transition matrix defining the measure prior at time :math:`t = n`.
        var_meas (ndarray(n_meas, n_meas)): Variance matrix defining the measure prior at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_pred** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        - **var_state_pred** (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        - **mean_state_filt** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        - **var_state_filt** (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
    """

    mean_state_pred, var_state_pred = predict(
        mean_state_past=mean_state_past,
        var_state_past=var_state_past,
        mean_state=mean_state,
        wgt_state=wgt_state,
        var_state=var_state
    )
    mean_state_filt, var_state_filt = update(
        mean_state_pred=mean_state_pred,
        var_state_pred=var_state_pred,
        x_meas=x_meas,
        mean_meas=mean_meas,
        wgt_meas=wgt_meas,
        var_meas=var_meas
    )
    return mean_state_pred, var_state_pred, mean_state_filt, var_state_filt


def _smooth(var_state_filt, var_state_pred, wgt_state):
    r"""
    Common part of :func:`smooth_sim` and :func:`smooth_mv`.

    Args:
        var_state_filt (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.

    Returns:
        (tuple):
        - **var_state_temp** (ndarray(n_state, n_state)): Temporary variance calculation used by :func:`smooth_sim`.
        - **var_state_temp_tilde** (ndarray(n_state, n_state)): Temporary variance calculation used by both :func:`smooth_sim` and :func:`smooth_mv`.
    """

    var_state_temp = var_state_filt.dot(wgt_state.T)
    var_state_temp_tilde = solve_var(var_state_pred, var_state_temp.T).T
    return var_state_temp, var_state_temp_tilde


def smooth_mv(mean_state_next,
              var_state_next,
              mean_state_filt,
              var_state_filt,
              mean_state_pred,
              var_state_pred,
              wgt_state,
              *args, **kwargs):
    r"""
    Performs one step of the Kalman mean/variance smoother.

    Calculates the mean and variance of :math:`p(X_n | Z_{0:N})`.

    Args:
        mean_state_next (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        var_state_next (ndarray(n_state, n_state)): State variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        mean_state_filt (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_filt (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_smooth** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_smooth** (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
    """

    var_state_temp, var_state_temp_tilde = _smooth(
        var_state_filt, var_state_pred, wgt_state
    )
    mean_state_smooth = mean_state_filt + \
        var_state_temp_tilde.dot(mean_state_next - mean_state_pred)
    var_state_smooth = var_state_filt + jnp.linalg.multi_dot(
        [var_state_temp_tilde, (var_state_next - var_state_pred), var_state_temp_tilde.T])
    return mean_state_smooth, var_state_smooth


def smooth_sim(x_state_next,
               mean_state_filt,
               var_state_filt,
               mean_state_pred,
               var_state_pred,
               wgt_state,
               *args, **kwargs):
    r"""
    Performs one step of the Kalman sampling smoother.

    Calculates the mean and variance of :math:`p(X_n | X_{n+1}, Z_{0:N})`.

    Args:
        x_state_next (ndarray(n_state)): Simulated state at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        mean_state_filt (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_filt (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_sim** (ndarray(n_state)): Mean for the sample solution at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_sim** (ndarray(n_state)): Variance for the sample solution at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
    """

    var_state_temp, var_state_temp_tilde = _smooth(
        var_state_filt, var_state_pred, wgt_state
    )
    mean_state_sim = mean_state_filt + \
        var_state_temp_tilde.dot(x_state_next - mean_state_pred)
    var_state_sim = var_state_filt - \
        var_state_temp_tilde.dot(var_state_temp.T)
    return mean_state_sim, var_state_sim


def smooth(x_state_next,
           mean_state_next,
           var_state_next,
           mean_state_filt,
           var_state_filt,
           mean_state_pred,
           var_state_pred,
           wgt_state,
           *args, **kwargs):
    r"""
    Performs one step of both Kalman mean/variance and sampling smoothers.

    Args:
        x_state_next (ndarray(n_state)): Simulated state at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        mean_state_next (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        var_state_next (ndarray(n_state, n_state)): State variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        mean_state_filt (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_filt (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_sim** (ndarray(n_state)): Mean for the sample solution at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_sim** (ndarray(n_state)): Variance for the sample solution at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **mean_state_smooth** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_smooth** (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
    """

    var_state_temp, var_state_temp_tilde = _smooth(
        var_state_filt, var_state_pred, wgt_state
    )
    mean_state_temp = jnp.concatenate([x_state_next[None],
                                       mean_state_next[None]])
    mean_state_temp = mean_state_filt + \
        var_state_temp_tilde.dot((mean_state_temp - mean_state_pred).T).T
    mean_state_sim = mean_state_temp[0]
    var_state_sim = var_state_filt - \
        var_state_temp_tilde.dot(var_state_temp.T)
    mean_state_smooth = mean_state_temp[1]
    var_state_smooth = var_state_filt + \
        jnp.linalg.multi_dot(
            [var_state_temp_tilde, (var_state_next - var_state_pred),
             var_state_temp_tilde.T])
    return mean_state_sim, var_state_sim, mean_state_smooth, var_state_smooth,


def forecast(mean_state_pred,
             var_state_pred,
             mean_meas,
             wgt_meas,
             var_meas,
             *args, **kwargs):
    r"""
    Forecasts the mean and variance of the measurement at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.

    Args:
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        var_state_pred (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        mean_meas (ndarray(n_meas)): Measurement equation offset at time :math:`t = n`.
        wgt_meas (ndarray(n_meas, n_state)): Measurement transition matrix at time :math:`t = n`.
        var_meas (ndarray(n_meas, n_meas)): State variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_fore** (ndarray(n_meas)): Forecast mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        - **var_fore** (ndarray(n_meas, n_meas)): Forecast variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
    """

    # wgt_meas = W + wgt_meas
    mean_fore = wgt_meas.dot(mean_state_pred) + mean_meas
    var_fore = jnp.linalg.multi_dot(
        [wgt_meas, var_state_pred, wgt_meas.T]) + var_meas
    return mean_fore, var_fore


def smooth_cond(mean_state_filt,
                var_state_filt,
                mean_state_pred,
                var_state_pred,
                wgt_state,
                *args, **kwargs):
    r"""
    Performs one step of the Kalman sampling smoother conditional.

    Finds :math:`A_n`, :math:`b_n`, and :math:`V_n` such that :math:`X_n \sim \mathcal{N}(A_n X_{n+1} + b_n, V_n)`. This is similar to the Kalman sampling algorithm but without the explicit sample.

    Args:
        mean_state_filt (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_filt (ndarray(n_state, n_state)): State variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **wgt_state_cond** (ndarray(n_state, n_state)): Transition of smooth conditional at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **mean_state_cond** (ndarray(n_state)): Offset of smooth conditional at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_cond** (ndarray(n_state, n_state)): Variance of smooth conditional at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
    """

    var_state_temp, wgt_state_cond = _smooth(
        var_state_filt, var_state_pred, wgt_state
    )
    mean_state_cond = mean_state_filt - wgt_state_cond.dot(mean_state_pred)
    var_state_cond = var_state_filt - wgt_state_cond.dot(var_state_temp.T)
    return wgt_state_cond, mean_state_cond, var_state_cond




r"""
Computes the initial parameters for the process prior using the q times integrated Brownian motion (IBM)

.. math::

    x^q(t) = \sigma B(t).

It has the analytical formulas for the parameters

.. math::

    Q_{ij} = 𝟙_{i\leq j}\frac{(\Delta t)^{j-i}}{(j-i)!}, \qquad R_{ij} = \sigma^2 \frac{(\Delta t)^{2q-1-i-j}}{(2q-1-i-j)(q-i)!(q-j)!}.

"""


import jax
import jax.numpy as jnp
import jax.scipy as jsp

def _factorial(x):
    """
    JAX factorial function.

    It's actually the gamma function shifted such that `_factorial(x) = x!` for integer-valued `x`.

    Args:
        x (int): Integer.
    
    Returns:
        (int): Factorial of x.

    """
    return jnp.exp(jsp.special.gammaln(x+1.0))


def ibm_state(dt, q, sigma):
    """
    Calculate the state weight matrix and variance matrix of q-times integrated Brownian motion.

    Args:
        dt (float): The step size between simulation points.
        q (int): The number of times to integrate the underlying Brownian motion.
        sigma (float): Parameter in the q-times integrated Brownian Motion.

    Returns:
        (tuple):
        - **Q** (ndarray(q+1, q+1)): The state weight matrix defined in
          Kalman solver.
        - **R** (ndarray(q+1, q+1)): The state variance matrix defined in
          Kalman solver.

    """
    I, J = jnp.meshgrid(jnp.arange(q+1), jnp.arange(q+1),
                        indexing="ij", sparse=True)
    mesh = J-I
    Q = jnp.nan_to_num(dt**mesh/_factorial(mesh), 0)
    mesh = (2.0*q+1.0) - I - J
    num = dt**mesh
    den = mesh * _factorial(q - I) * _factorial(q-J)
    R = sigma**2 * num/den
    return Q, R


def ibm_init(dt, n_deriv, sigma):
    """
    Calculates the initial parameters necessary for the Kalman solver with the p-1 times
    integrated Brownian Motion, where the ODE is up to the p-1th derivative.

    Args:
        dt (float): The step size between simulation points.
        n_deriv (int): Dimension of the prior.
        sigma (ndarray(n_block)): Parameter in variance matrix.

    Returns:
        (tuple):
        - **wgt_state** (ndarray(n_block, p, p)): Weight matrix defining the solution prior; :math:`Q`.
        - **var_state** (ndarray(n_block, p, p)): Variance matrix defining the solution prior; :math:`R`.

    """
    n_block = len(sigma)

    wgt_state_i, var_state_i = ibm_state(dt, n_deriv-1, 1)
    wgt_state = jnp.repeat(wgt_state_i[None,:,:], repeats=n_block, axis=0)
    var_state = jax.vmap(lambda b:
        sigma[b]**2 * var_state_i)(jnp.arange(n_block))
    
    return wgt_state, var_state



import unittest
from scipy.integrate import odeint


from rodeo import solve_sim, solve_mv
from rodeo.interrogate import interrogate_rodeo
import utils

class TestFitzOdeint(unittest.TestCase):
    """
    Check whether rodeo and odeint gives approximately the same results.
    
    """
    setUp = utils.fitz_setup

    def test_fitz(self):
        det = odeint(self.fitz_odeint, self.x0, self.tseq, args=(self.theta,))
        sim = solve_sim(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                        ode_init=self.x0_block, theta=self.theta,
                        t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                        prior_pars=self.prior_pars,
                        interrogate=interrogate_rodeo)
        m = solve_mv(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                     ode_init=self.x0_block, theta=self.theta,
                     t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                     prior_pars=self.prior_pars,
                     interrogate=interrogate_rodeo)[0]
        self.assertLessEqual(utils.rel_err(sim[:, :, 0], det), 5.0)
        self.assertLessEqual(utils.rel_err(m[:, :, 0], det), 5.0)

if __name__ == '__main__':
    unittest.main()



r"""
This module implements the Basic method for computing the approximate loglikelihood of :math:`\log p(Y_{0:M} \mid Z_{1:N})`.

Using :math:`\mu_{0:N|N} = E(X_{0:N} \mid Z_{1:N})` from the rodeo solver, the approximate likelihood is computed as

.. math:: p(Y_{0:M} \mid Z_{1:N}) = \sum_{i=0}^M \log p(Y_i \mid X_{n(i)} = \mu_{n(i)|N}).

In the case that observations time grid is not the same as the solver time grid, then the observation uses the closest discretization time point.

"""
import jax
import jax.numpy as jnp
from rodeo.solve import solve_mv


def basic(key, ode_fun, ode_weight, ode_init, 
          t_min, t_max, n_steps,
          interrogate,
          prior_pars,
          obs_data, obs_times, obs_loglik,
          kalman_type="standard", **params):
    
    r"""
    Basic algorithm to compute the approximate loglikelihood of :math:`p(Y_{0:M} \mid Z_{1:N})`.

    Args:
        key (PRNGKey): PRNG key.
        ode_fun (Callable): Higher order ODE function :math:`W X_t = F(X_t, t)` taking arguments :math:`X` and :math:`t`.
        ode_weight (ndarray(n_block, n_bmeas, n_bstate)): Weight matrix defining the measure prior; :math:`W`.
        ode_init (ndarray(n_block, n_bstate)): Initial value of the state variable :math:`X_t` at time :math:`t = a`.
        t_min (float): First time point of the time interval to be evaluated; :math:`a`.
        t_max (float): Last time point of the time interval to be evaluated; :math:`b`.
        n_steps (int): Number of discretization points (:math:`N`) of the time interval that is evaluated, such that discretization timestep is :math:`dt = (b-a)/N`.
        interrogate (Callable): Function defining the interrogation method.
        prior_pars (tuple): A tuple containing the weight matrix and the variance matrix defining the solution prior; :math:`Q, R`.
        obs_data (ndarray(n_obs, n_bobs)): Observed data; :math:`Y_{0:M}`.
        obs_times (ndarray(n_obs)): Observation time; :math:`0, \ldots, M`.
        obs_loglik (Callable): Observation loglikelihood function.
        kalman_type (str): Determine which type of Kalman (standard, square-root) to use.
        params (kwargs): Optional model parameters.

    Returns:
        (float): The loglikelihood of :math:`p(Y_{0:M} \mid Z_{1:N})`.

    """

    Xt, _ = solve_mv(
        key=key,
        ode_fun=ode_fun,
        ode_weight=ode_weight,
        ode_init=ode_init,
        t_min=t_min,
        t_max=t_max,
        n_steps=n_steps,
        interrogate=interrogate,
        prior_pars=prior_pars,
        kalman_type=kalman_type,
        **params
    )
    sim_times = jnp.linspace(t_min, t_max, n_steps+1)
    ode_data = Xt[jnp.searchsorted(sim_times, obs_times)]
    return obs_loglik(obs_data, ode_data, **params), Xt
